from collections import defaultdict
import re
import string
import json
from tqdm import tqdm
import numpy as np
import pandas as pd
from typing import List, Dict, Tuple
import torch
import random
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import linear_kernel
from .file_utils import is_scatter_available
from tatqa_utils import *
from .data_util import *
from .data_util import _is_average, _is_change_ratio, _is_diff, _is_division, _is_sum, _is_times
from nltk.tokenize import sent_tokenize, word_tokenize
import re
import dgl
from .pre_order import infix_to_prefix
from .dataset_util import  change_output2
from reg_hnt.reghnt.Vocabulary import *
from collections import defaultdict, OrderedDict
from itertools import groupby
from nltk.corpus import stopwords

stop_words = stopwords.words('english')

def get_order_by_tf_idf(question, paragraphs):
    sorted_order = []
    corpus = [question]
    for order, text in paragraphs.items():
        corpus.append(text)
        sorted_order.append(order)
    tf_idf = TfidfVectorizer().fit_transform(corpus)
    cosine_similarities = linear_kernel(tf_idf[0:1], tf_idf).flatten()[1:]
    sorted_similarities = sorted(enumerate(cosine_similarities),
                                 key=lambda x: x[1])
    idx = [i[0] for i in sorted_similarities][::-1]
    return [sorted_order[index] for index in idx]




def is_whitespace(c):
    if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
        return True
    return False


def is_year(item):
    years = [str(i) for i in range(2001, 2051)]
    for year in years:
        if year in item:
            return True
    return False


def is_month(item):
    months = ['January','February','August','March',\
        'April','May','June','July','September',\
            'October','November','December']
    for month in months:
        if month in item:
            return True
    return False


def is_scale(item):
    item = sep_lower(item)
    scales = [
        '4rxr7y99', 'thousand', 'million', 'billion', 'percent', '%', '£m',
        'USDm', 'EURm', '€m', '$m', '$M', 'US$’000', '$’000', "£’000", '£000',
        'US$000', '$000', "$'000"
    ]
    for scale in scales:
        if scale in item:
            return True
    return False


def is_None(item):
    if item == '' or item == 'N/A' or item == 'n/a' or item == '—' or item == '–' or item == '~'\
    or item == '$—' or item == '—%'or item =='$ —' or item == '— %' or item == "-" or item=='(%)'\
    or item == "$-"or item == "$ -" or item=="-%" or item=="- %"  or item == 'nm' or item=='nm ' or item=='n.m.'\
    or item=="*" or item=="$'000" or item=="$" or item=="%" or item == 'n/m' or item =='$’000'\
    or item=='Years' or item=='£m' or item=='USDm' or item=="EURm" or item == '€m' or item == '$m' or item=='$M'\
    or item=='US$’000' or item=="£’000" or item=='£000' or item=='$000' or item=='US$000' or item=='---':
        return True
    else:
        return False


# answer_dict = {"answer_type": answer_type, "answer": answer, "scale": scale, "answer_from": answer_from}
def generate_class(answer_dict, mapping):
    operator_class = -1
    if answer_dict['answer_type'] == "span":
        operator_class = OPERATOR_CLASSES["SPAN"]
    elif answer_dict['answer_type'] == "multi-span":
        operator_class = OPERATOR_CLASSES["MULTI_SPAN"]
    elif answer_dict['answer_type'] == "count":
        operator_class = OPERATOR_CLASSES["COUNT"]
    elif answer_dict['answer_type'] == "arithmetic":
        operator_class = OPERATOR_CLASSES["ARITHMETIC"]
    return operator_class


def string_tokenizer(string: str, tokenizer) -> List[int]:
    if not string:
        return []
    tokens = []
    string_subword_lens = []
    string_types = []
    prev_is_whitespace = True
    for i, c in enumerate(string):
        if is_whitespace(c):  # or c in ["-", "–", "~"]:
            prev_is_whitespace = True
        else:
            if prev_is_whitespace:
                tokens.append(c)
            else:
                tokens[-1] += c
            prev_is_whitespace = False  # char_to_word_offset.append(len(tokens) - 1)
    split_tokens = []
    for i, token in enumerate(tokens):
        if i != 0:
            sub_tokens = tokenizer.tokenize(" " + token)
        else:
            sub_tokens = tokenizer.tokenize(token)

        for sub_token in sub_tokens:
            split_tokens.append(sub_token)
        string_subword_lens += [len(sub_tokens)]
    ids = tokenizer.convert_tokens_to_ids(split_tokens)
    return ids, string_subword_lens, tokens




def question_tokenize(question, tokenizer, mapping):
    if not question:
        return []
    tokens = []

    question_word_index = []
    question_subword_lens = []
    question_types = []
    question_word_types = []
    current_index = 1
    prev_is_whitespace = True
    question = question.replace("?", '')
    question_mapping, question_answer_index = False, []
    question_number_value = []

    if 'question' in list(mapping.keys()):
        question_mapping = True
        question_answer_index = mapping['question']

    current_tags = [0 for i in range(len(question))]
    try:
        if question_mapping:
            for answer_index in question_answer_index:
                current_tags[answer_index[0]:answer_index[1]] = \
                    [1 for i in range(len(current_tags[answer_index[0]:answer_index[1]]))]
    except:
        import pdb; pdb.set_trace()
    tags = []
    wait_add = False
    start_index = 0

    for i, c in enumerate(question):
        if is_whitespace(c):  # or c in ["-", "–", "~"]:
            if wait_add:
                if 1 in current_tags[start_index:i]:
                    tags.append(1)
                else:
                    tags.append(0)
                wait_add = False
            prev_is_whitespace = True
        else:
            if prev_is_whitespace:
                tokens.append(c)
                wait_add = True
                start_index = i
            else:
                tokens[-1] += c
            prev_is_whitespace = False  # char_to_word_offset.append(len(tokens) - 1)
    if wait_add:
        if 1 in current_tags[start_index:len(question)]:
            tags.append(1)
        else:
            tags.append(0)
    split_tokens = []
    # if question_mapping:
    #     import pdb; pdb.set_trace()
    for i, token in enumerate(tokens):
        if i != 0:
            sub_tokens = tokenizer.tokenize(" " + token)
        else:
            sub_tokens = tokenizer.tokenize(token)

        flag = 1
        subword_length = 0
        for sub_token in sub_tokens:
            split_tokens.append(sub_token)
            question_word_index.append(current_index)
            flag = 0
            subword_length += 1

        question_subword_lens += [subword_length]
        current_index += 1

        if is_number(token) and not is_year(token) and not is_month(
                tokens[i - 1]):
            question_number_value.append(float(to_number(token)))
            question_types += [2 for _ in range(len(sub_tokens))]
            question_word_types += [2]
        else:
            question_number_value.append(np.nan)
            question_types += [3 for _ in range(len(sub_tokens))]
            question_word_types += [3]

    ids = tokenizer.convert_tokens_to_ids(split_tokens)

    return ids, tokens, question_subword_lens, question_word_index, question_types, question_word_types, question_number_value, tags




def clean_year(item):
    flag = 0
    years = [str(i) for i in range(2000, 2051)]
    num = 0
    this_year = ''
    for year in years:
        if year in item:
            num += 1
            this_year = year
    if num != 1 or item == '2015/16' or item == '2016/17' or item == '2017/18' or item == '2018/19' or item == '2019/20' or item == "2025-beyond":
        return item
    else:
        return this_year


def scale_judge(item):
    item = sep_lower(item)
    scales = [
        '4rxr7y99', 'thousand', 'million', 'billion', 'percent', '%', '£m',
        'USDm', 'EURm', '€m', '$m', '$M', 'US$’000', '$’000', "£’000", '£000',
        'US$000', '$000', "$'000"
    ]
    for idx, scale in enumerate(scales):
        if scale in item:
            to_idx = 0
            if idx in [1, 12, 13, 14, 15, 16, 17, 18]:
                to_idx = 1
            elif idx in [2, 6, 7, 8, 9, 10, 11]:
                to_idx = 2
            elif idx in [3]:
                to_idx = 3
            elif idx in [4, 5]:
                to_idx = 4
            return idx, to_idx


def generate_number_scale(table, table_pd, table_cell_type_ids,table_is_scale_type, table_cell_number_value,table_cell_number_value_position, paragraphs):
    SCALE = ['', 'thousand', 'million', 'billion', 'percent']
    TMB_SCALE = ['thousand', 'million', 'billion']

    table_number_position = []
    for x, y in zip(table_cell_number_value_position, table_cell_number_value):
        if not np.isnan(y):
            table_number_position.append(x)
        else:
            table_number_position.append((-1, -1))

    table_cell_number_scale = list(
        map(lambda x: 0 if not np.isnan(x) else x, table_cell_number_value))
    global_scale = 0

    scale_position = np.where(table_is_scale_type > 0)
    for x, y in zip(*scale_position):

        this_scale, to_scale = scale_judge(table[x][y])
        if this_scale in [1, 2, 3, 6, 7]:
            global_scale = to_scale

            for idx in range(x, len(table)):
                for idy in range(len(table[idx])):
                    if (idx, idy) in table_cell_number_value_position:
                        table_cell_number_scale[table_cell_number_value_position.index((idx, idy))] = to_scale

    if global_scale == 0:
        for paragraph in paragraphs:
            paragraph_text = paragraph['text']
            if 'in thousand' in paragraph_text or 'In thousands' in paragraph_text:
                global_scale = 1
            if 'in million' in paragraph_text or 'In million' in paragraph_text:
                global_scale = 2
            if 'in billion' in paragraph_text or 'In billion' in paragraph_text:
                global_scale = 3
        if global_scale != 0:
            table_cell_number_scale = list(map(lambda x: global_scale if not np.isnan(x) else x, table_cell_number_scale))

    for x, y in zip(*scale_position):
        this_scale, to_scale = scale_judge(table[x][y])
        if this_scale in [4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]:
            for item in table_number_position:
                if item[1] == y and item[0] > x:
                    table_cell_number_scale[table_cell_number_value_position.index(item)] = to_scale
                if item[0] == x and item[1] > y:
                    table_cell_number_scale[table_cell_number_value_position.index(item)] = to_scale

    for i in range(len(table)):
        for j in range(len(table[i])):
            if table_cell_type_ids[i][j] == 2:
                if '%' in table[i][j]:
                    table_cell_number_scale[table_cell_number_value_position.index((i, j))] = 4

    return table_cell_number_scale


def table_tokenize(table, tokenizer, mapping, question_id):
    table_mapping = False
    if "table" in mapping and len(mapping["table"]) != 0:
        table_mapping = True
        answer_coordinates = mapping["table"]
    table_pd = pd.DataFrame(table, dtype=np.str)
    table_cell_type_ids = np.full(table_pd.shape, -1)
    table_is_scale_type = np.zeros(table_pd.shape)
    table_cell_tokens = []
    table_cell_number_value = []
    table_cell_number_value_position = []
    table_cell_number_value_position_b = []
    table_ids = []
    table_tags = []
    table_cell_index = []
    table_subword_lens = []
    table_types = []
    table_word_types = []
    current_cell_index = 0
    table_word_tokens = []
    table_word_tokens_number = []
    table_cell_types_flatten = []

    row_dict = defaultdict(list)
    column_dict = defaultdict(list)

    for i in range(len(table)):
        for j in range(len(table[i])):

            table[i][j] = table[i][j].replace('bps', '')

            if table[i][j] in ['$—', '—%', '$ —', '— %', "-","—","$-","$ -","-%","- %"]:
                table[i][j] = '0'

            if is_None(table[i][j]):
                table_cell_type_ids[i][j] = 0
            elif is_number(table[i][j]) and not is_year(table[i][j]):
                table_cell_type_ids[i][j] = 2
            elif is_year(table[i][j]):
                table_cell_type_ids[i][j] = 1
            else:
                flag1 = 0
                for item in table[i]:
                    if item != '':
                        flag1 += 1

                flag = 0
                for item in table[i]:
                    if is_number(item) and not is_year(item):
                        flag = 1
                if flag:
                    table_cell_type_ids[i][j] = 3
                flag = 0
                try:
                    for item in [item[j] for item in table]:
                        if is_number(item) and not is_year(item):
                            flag = 1
                except:
                    import pdb
                    pdb.set_trace()
                if flag and flag1 != 1:
                    table_cell_type_ids[i][j] = 4

            if is_scale(table[i][j]) and not is_number(table[i][j]):
                table_is_scale_type[i][j] = 1

            if table_cell_type_ids[i][j] == 0:
                continue

            table_cell_input_token = table[i][j]

            table_cell_tokens.append(table[i][j])

            if table_cell_type_ids[i][j] == 2:
                per_flag = 0
                for c_idx in range(len([item[j] for item in table])):
                    if table_is_scale_type[c_idx][j] == 1 and scale_judge(table[c_idx][j]) == 4:
                        per_flag = 1
                for c_idx in range(len(table[i])):
                    if table_is_scale_type[i][c_idx] == 1 and scale_judge(table[i][c_idx]) == 4:
                        per_flag = 1
                num_number = to_number(table[i][j])
                if per_flag and '%' not in table[i][j]:
                    num_number = round(float(num_number / 100.0), 4)
                table_cell_number_value.append(num_number)
                table_cell_input_token = table[i][j].replace(' ', '').replace('$', '')
                if '(' in table_cell_input_token and ')' in table_cell_input_token:
                    table_cell_input_token = table_cell_input_token.replace('(', '').replace(')', '')
                    table_cell_input_token = '-' + table_cell_input_token
                table_cell_number_value_position.append((i, j))
            else:
                table_cell_number_value.append(np.nan)
                table_cell_number_value_position.append((-1, -1))
            cell_ids, string_subword_lens, tokens = string_tokenizer(table_cell_input_token, tokenizer)
            table_word_tokens.extend(tokens)
            table_word_tokens_number += [len(tokens)]
            table_subword_lens += string_subword_lens

            if table_cell_type_ids[i][j] == 2:
                table_types += [2 for _ in range(len(cell_ids))]
                table_word_types += [2 for _ in range(len(string_subword_lens))]
            else:
                table_types += [3 for _ in range(len(cell_ids))]
                table_word_types += [3 for _ in range(len(string_subword_lens))]

            table_ids += cell_ids
            table_cell_index += [current_cell_index for _ in range(len(cell_ids))]
            if table_mapping:
                if [i, j] in answer_coordinates:
                    table_tags += [1]
                else:
                    table_tags += [0]
            else:
                table_tags += [0]

            row_dict[i].append(current_cell_index)
            column_dict[j].append(current_cell_index)
            table_cell_types_flatten.append(table_cell_type_ids[i][j])
            current_cell_index += 1
    return table_pd,table_cell_type_ids,table_is_scale_type,table_cell_tokens,table_cell_number_value,\
        table_ids,table_tags,table_cell_index, table_subword_lens, table_types, table_word_types,table_word_tokens,table_word_tokens_number,table_cell_number_value_position,\
            row_dict, column_dict, table_cell_types_flatten


def paragraph_number_judge(tokens, idx, this_number, paragraphs):
    global_scale = 0

    for i, paragraph_text in paragraphs.items():
        if 'in thousand' in paragraph_text:
            global_scale = 1
        if 'in million' in paragraph_text:
            global_scale = 2
        if 'in billion' in paragraph_text:
            global_scale = 3

    flag = 1
    scale_flag = 0
    if idx != 0:
        if is_month(tokens[idx - 1]) and '$' not in this_number:
            flag = 0
    if idx != len(tokens) - 1:
        if is_month(tokens[idx + 1]) and '$' not in this_number:
            flag = 0
    if idx == 0:
        flag = 0

    if this_number[
            -1] == '.' and '$' not in this_number and '%' not in this_number and '£' not in this_number:
        flag = 0

    for stopword in ['(1)', '(2)', '(3)', '(4)', '(5)', '(6)', '(7)']:
        if stopword in this_number:
            flag = 0
            break

    if flag != 0:
        if '%' in this_number:
            scale_flag = 4

        if 'm' in this_number:
            scale_flag = 2
        if 'b' in this_number:
            scale_flag = 3

        for token in tokens[idx + 1:idx + 3]:
            token = sep_lower(token)
            for scale_idx, scale_label in enumerate(['dfe526f2', 'thousand', 'million', 'billion', 'percent']):
                if scale_label in token:
                    scale_flag = scale_idx

    if scale_flag == 0 and global_scale != 0:
        scale_flag = global_scale

    return flag, scale_flag


def paragraph_tokenize(question, paragraphs, tokenizer, mapping, question_id):
    paragraphs_copy = paragraphs.copy()
    paragraphs = {}
    for paragraph in paragraphs_copy:
        paragraphs[paragraph["order"]] = paragraph["text"]
    del paragraphs_copy
    split_tokens = []
    split_tags = []
    number_mask = []
    number_value = []

    tokens = []
    tags = []
    paragraph_subword_tags = []
    paragraph_index = []
    paragraph_tags = []

    paragraph_subword_lens = []
    paragraph_types = []
    paragraph_word_types = []
    paragraph_number_scale = []
    paragraph_sep_tag = []
    paragraph_mapping = False
    paragraph_mapping_orders = []
    try:
        if "paragraph" in list(
                mapping.keys()) and len(mapping["paragraph"].keys()) != 0:
            paragraph_mapping = True
            paragraph_mapping_orders = list(mapping["paragraph"].keys())
    except:
        import pdb;pdb.set_trace()

    sorted_order = get_order_by_tf_idf(question, paragraphs)

    for order in sorted_order:
        try:
            text = paragraphs[order]
        except:
            import pdb; pdb.set_trace()
        text = " ".join(text.split())
        prev_is_whitespace = True
        answer_indexs = None
        if paragraph_mapping and str(order) in paragraph_mapping_orders:
            answer_indexs = mapping["paragraph"][str(order)]
        current_tags = [0 for i in range(len(text))]
        if answer_indexs is not None:
            for answer_index in answer_indexs:
                current_tags[answer_index[0]:answer_index[1]] = \
                    [1 for i in range(len(current_tags[answer_index[0]:answer_index[1]]))]

        start_index = 0
        wait_add = False

        tokens.append(tokenizer.sep_token)
        tags.append(0)
        for i, c in enumerate(text):
            if is_whitespace(c):  # or c in ["-", "–", "~"]:
                if wait_add:
                    if 1 in current_tags[start_index:i]:
                        tags.append(1)
                    else:
                        tags.append(0)
                    wait_add = False
                prev_is_whitespace = True
            else:
                if prev_is_whitespace:
                    tokens.append(c)
                    wait_add = True
                    start_index = i
                else:
                    tokens[-1] += c
                prev_is_whitespace = False
        if wait_add:
            if 1 in current_tags[start_index:len(text)]:
                tags.append(1)
            else:
                tags.append(0)

    if "direct‐to‐grower" in tokens:
        tokens.remove("direct‐to‐grower")
    if "ISR-as-a-service" in tokens:
        tokens.remove("ISR-as-a-service")

    current_token_index = 0
    for i, token in enumerate(tokens):
        if token == tokenizer.sep_token:
            sub_tokens = tokenizer.tokenize(token)
            paragraph_sep_tag += [1]
        elif i != 0:
            sub_tokens = tokenizer.tokenize(" " + token)
            paragraph_sep_tag += [0]
        else:
            sub_tokens = tokenizer.tokenize(token)
            paragraph_sep_tag += [0]
        paragraph_subword_lens += [len(sub_tokens)]
        stop_flag = 0
        if is_number(token) and not is_year(token):
            stop_flag, scale_flag = paragraph_number_judge(tokens, i, token, paragraphs)
        try:
            if stop_flag == 1 and is_number(token) and not is_year(token):
                number_value.append(float(to_number(token)))
                paragraph_types += [2 for _ in range(len(sub_tokens))]
                paragraph_word_types += [2]
                paragraph_number_scale.append(scale_flag)
            else:
                number_value.append(np.nan)
                paragraph_types += [3 for _ in range(len(sub_tokens))]
                paragraph_word_types += [3]
                paragraph_number_scale.append(np.nan)
        except:
            import pdb; pdb.set_trace()
        flag = 0
        for sub_token in sub_tokens:
            if tags[i] == 1:
                flag = 1
            split_tags.append(tags[i])
            split_tokens.append(sub_token)
            paragraph_index.append(current_token_index)
        if flag:
            paragraph_tags.append(1)
        else:
            paragraph_tags.append(0)
        current_token_index += 1
        paragraph_subword_tags += [1]
        if len(sub_tokens) > 1:
            paragraph_subword_tags += [0] * (len(sub_tokens) - 1)
    paragraph_ids = tokenizer.convert_tokens_to_ids(split_tokens)
    return tokens, paragraph_ids, paragraph_tags, number_value, paragraph_index, paragraph_subword_lens, paragraph_types, paragraph_word_types, paragraph_number_scale, paragraph_sep_tag

def subword_judge(subword_lens, input_types):
    aa = len(input_types) - len(torch.where(torch.tensor(input_types) == 0)[0])
    bb = 0
    for i in subword_lens:
        bb += i
    if aa == bb:
        return subword_lens
    else:
        cc = bb - aa
        subword_lens[-1] -= cc
        return subword_lens


def _concat(question_ids, table_ids, paragraph_ids, table_tags, paragraph_tags,
            question_tokens, table_tokens, paragraph_tokens,
            table_cell_number_value, paragraph_number_value, table_cell_index,
            paragraph_index, question_subword_lens, question_types,
            table_subword_lens, table_types, paragraph_subword_lens,
            paragraph_types, question_word_types, table_word_types,
            paragraph_word_types, table_cell_number_value_position,
            table_cell_number_scale, paragraph_number_scale,
            question_number_value, question_tags, row_dict, column_dict,
            table_cell_types_flatten, paragraph_sep_tag, sep,
            question_length_limitation, passage_length_limitation, max_pieces,
            question_id, tokenizer):

    input_ids = torch.zeros([1, max_pieces])
    token_type_ids = torch.zeros_like(input_ids)
    question_ids = [sep] + question_ids + [sep]
    question_types = [0] + question_types + [0]
    question_length = len(question_ids)
    table_length = len(table_ids)
    paragraph_length = len(paragraph_ids)
    table_cell_count_origin = len(table_cell_types_flatten)
    if passage_length_limitation is not None:
        passage_length_limitation = passage_length_limitation - question_length - 2
        if len(table_ids) > passage_length_limitation:
            passage_ids = table_ids[:passage_length_limitation]
            table_length = passage_length_limitation
            paragraph_length = 0
            table_cell_index = table_cell_index[:passage_length_limitation]
            table_types = table_types[:passage_length_limitation]
            end = table_cell_index[-1]
            table_subword_lens = table_subword_lens[:end + 1]
            table_word_types = table_word_types[:end + 1]
            table_tags = table_tags[:end + 1]
            table_tokens = table_tokens[:end + 1]
            table_cell_types_flatten = table_cell_types_flatten[:end + 1]
            table_cell_number_value = table_cell_number_value[:end + 1]
            table_cell_number_value_position = table_cell_number_value_position[:end+1]
            table_cell_number_scale = table_cell_number_scale[:end + 1]
        elif len(table_ids) + len(paragraph_ids) > passage_length_limitation:
            passage_ids = table_ids + [sep] + paragraph_ids
            passage_ids = passage_ids[:passage_length_limitation]
            paragraph_index = paragraph_index[:passage_length_limitation - table_length - 1]
            paragraph_types = paragraph_types[:passage_length_limitation - table_length - 1]
            end = paragraph_index[-1]
            paragraph_tags = paragraph_tags[:end + 1]
            paragraph_tokens = paragraph_tokens[:end + 1]
            paragraph_sep_tag = paragraph_sep_tag[:end + 1]
            paragraph_number_value = paragraph_number_value[:end + 1]
            paragraph_number_scale = paragraph_number_scale[:end + 1]
            paragraph_subword_lens = paragraph_subword_lens[:end + 1]
            paragraph_word_types = paragraph_word_types[:end + 1]
            table_length = len(table_ids)
            paragraph_length = passage_length_limitation - table_length
        else:
            passage_ids = table_ids + [sep] + paragraph_ids
            table_length = len(table_ids)
            paragraph_length = len(paragraph_ids) + 1
    else:
        passage_ids = table_ids + [sep] + paragraph_ids

    passage_ids = passage_ids + [sep]

    input_ids[0, :question_length] = torch.from_numpy(np.array(question_ids))
    input_ids[0, question_length:question_length +
              len(passage_ids)] = torch.from_numpy(np.array(passage_ids))
    attention_mask = input_ids != 0

    input_types = question_types + table_types
    if paragraph_length > 1:
        input_types += [0] + paragraph_types

    question_tags = np.zeros_like(question_tags).tolist()
    tags = torch.from_numpy(np.array(question_tags + table_tags + paragraph_tags))

    try:
        table_number = [item for item in table_cell_number_value if not np.isnan(item)]
    except:
        import pdb; pdb.set_trace()

    paragraph_number = [item for item in paragraph_number_value if not np.isnan(item)]
    question_number = [item for item in question_number_value if not np.isnan(item)]

    table_number_position = [x for x, y in zip(table_cell_number_value_position, table_cell_number_value) if not np.isnan(y)]

    table_number_index = [i for i, item in enumerate(table_cell_number_value) if not np.isnan(item)]
    paragraph_number_index = [i for i, item in enumerate(paragraph_number_value) if not np.isnan(item)]
    question_number_index = [i for i, item in enumerate(question_number_value) if not np.isnan(item)]

    table_number_scale = [item for i, item in enumerate(table_cell_number_scale) if not np.isnan(item)]
    paragraph_number_scale = [item for i, item in enumerate(paragraph_number_scale) if not np.isnan(item)]

    subword_lens = question_subword_lens + table_subword_lens + paragraph_subword_lens
    input_word_types = question_word_types + table_word_types + paragraph_word_types

    subword_lens = subword_judge(subword_lens, input_types)

    table_cell_count = len(table_cell_types_flatten)
    if table_cell_count_origin != table_cell_count:
        import pdb;pdb.set_trace()

    for key, key_list in row_dict.items():
        for item in key_list:
            if item >= table_cell_count:
                key_list.remove(item)
    for key, key_list in column_dict.items():
        for item in key_list:
            if item >= table_cell_count:
                key_list.remove(item)

##########################################################
    p_sep_fuzhu_tag = [0 for _ in range(paragraph_sep_tag.count(1))]
    p_sep_pos = [idx for idx, item in enumerate(paragraph_sep_tag) if item == 1]
    p_sep_span = []
    for idx in range(len(p_sep_pos)):
        if idx < len(p_sep_pos) - 1:
            p_sep_span.append((p_sep_pos[idx], p_sep_pos[idx + 1]))
        else:
            p_sep_span.append((p_sep_pos[idx], len(paragraph_sep_tag)))

    for idx, item in enumerate(p_sep_span):
        split_paragraph_tags = paragraph_tags[item[0]:item[1]]
        if 1 in split_paragraph_tags:
            p_sep_fuzhu_tag[idx] = 1

    answer_tp = [0, 0]
    if 1 in table_tags:
        answer_tp[0] = 1
    if 1 in paragraph_tags:
        answer_tp[1] = 1

    paragraph_sep_tag_new = [0 for _ in range(len(question_tokens) + len(table_tags) + len(paragraph_tags))]
    paragraph_sep_tag_new[len(question_tokens) + len(table_tags):] = paragraph_sep_tag

    sep_number = p_sep_fuzhu_tag.count(1)
    p_sep_fuzhu_tag_new = paragraph_sep_tag_new.copy()
    jilu = 0
    for i in range(len(p_sep_fuzhu_tag_new)):
        if p_sep_fuzhu_tag_new[i] == 1:
            if p_sep_fuzhu_tag[jilu] == 0:
                p_sep_fuzhu_tag_new[i] = 0
            jilu += 1


##############################################
    return input_ids, attention_mask, input_types, input_word_types, subword_lens, tags,\
        table_cell_index,table_tokens,table_number, table_number_index, paragraph_index,\
            paragraph_tokens, paragraph_number, paragraph_number_index,token_type_ids,table_number_position,\
                table_number_scale, paragraph_number_scale, question_number, question_number_index, row_dict, column_dict, table_cell_types_flatten, paragraph_sep_tag_new, p_sep_fuzhu_tag_new, answer_tp

OPERATOR = ['+', '-', '*', '/']


def get_operators(derivation: str):
    res = []
    for c in derivation:
        if c in OPERATOR:
            res.append(c)
    return res


from itertools import chain


def generate_output2(answer_type, derivation, question_number, table_number,
                     paragraph_number, question, table, paragraphs, mapping,
                     facts, table_number_position, answer, table_pd,
                     question_id, paragraph_tokens):
    if answer_type != "arithmetic":
        return []
    output = []
    num_fact = facts_to_nums(facts)
    try:
        pre_order = infix_to_prefix(derivation)
    except:
        import pdb; pdb.set_trace()

    question_mapping, table_mapping, paragraph_mapping = None, None, None
    if 'table' in mapping.keys():
        table_mapping = mapping['table']
        try:
            sel_t_num = [table[span[0]][span[1]] for span in table_mapping]
        except:
            import pdb
            pdb.set_trace()
    if 'paragraph' in mapping.keys():
        paragraph_mapping = mapping['paragraph']
    if 'question' in mapping.keys():
        question_mapping = mapping['question']
        sel_q_num = [question[span[0]:span[1]] for span in question_mapping]
    pre_order_tokens = pre_order.split(' ')
    numbers = CONST_LIST2 + OPERATOR_LIST + table_number + paragraph_number

    def number_search(item):
        tat_num = to_number(item)
        if tat_num is None:
            import pdb
            pdb.set_trace()
        paragraphs, table_number_position, question_id, pre_order, pre_order_tokens,  table_number, paragraph_number
        if table_mapping:
            for span in table_mapping:
                ori_cell = table[span[0]][span[1]]
                try:
                    ori_num = to_number(ori_cell)
                except:
                    import pdb;pdb.set_trace()
                    continue
                if ori_num is None:
                    continue
                if ori_num == tat_num or (ori_num + tat_num == 0 and tat_num != 2) or ori_num == round(float(tat_num / 100.0), 4) or ori_num + round(float(tat_num / 100.0), 4) == 0 or round(float(tat_num * 100.0), 4) == ori_num:
                    if (span[0], span[1]) not in table_number_position:
                        import pdb;pdb.set_trace()
                        continue
                    else:
                        return len(CONST_LIST2 + OPERATOR_LIST ) + table_number_position.index((span[0], span[1]))

        if paragraph_mapping:
            for i, num in enumerate(paragraph_number):
                if num == tat_num or num == round(float(tat_num / 100.0),4) or num + tat_num == 0:
                    return i + len(CONST_LIST2 + OPERATOR_LIST + table_number)

        if str(tat_num) in CONST_LIST2:
            return len(OPERATOR_LIST) + CONST_LIST2.index(str(tat_num))

    table_number_copy = table_number.copy()

    for i, number in enumerate(table_number):
        x, y = table_number_position[i][0], table_number_position[i][1]
        flag = 0
        if table_mapping is not None:
            for item in table_mapping:
                if x == item[0] and y == item[1]:
                    flag = 1
            if flag == 0:
                table_number_copy[i] = np.nan


    if pre_order_tokens[0] == '/' and pre_order_tokens[1] == '+' and (
            pre_order_tokens[-1] in CONST_LIST
            or pre_order_tokens[-1].replace('$', '').strip() in CONST_LIST):
        for i, token in enumerate(pre_order_tokens[:-1]):
            if token in OPERATOR_LIST:
                output.append(OPERATOR_LIST.index(token))
                continue
            index = number_search(token)
            if index is None:
                import pdb; pdb.set_trace()
            output.append(index)
        try:
            output.append(CONST_LIST2.index(pre_order_tokens[-1]) + len(OPERATOR_LIST))
        except:
            import pdb
            pdb.set_trace()
    else:
        for i, token in enumerate(pre_order_tokens):
            if token in OPERATOR_LIST:
                output.append(OPERATOR_LIST.index(token))
                continue
            index = number_search(token)
            if index is None:
                import pdb; pdb.set_trace()
            output.append(index)
        if output[-1] is None:
            output[-1] = CONST_LIST2.index(pre_order_tokens[-1]) + len(OPERATOR_LIST)

    output, pre_order_tokens = change_output2(output, pre_order_tokens,question_id)
    if len(output) % 2 == 0:
        import pdb;pdb.set_trace()
    if len(pre_order_tokens) != len(output):
        import pdb;pdb.set_trace()

    return output

def delete_punc(example):
    if example == '[SEP]' or example == '</s>' or example == '<s>':
        return example
    punctuation_string = string.punctuation
    for i in punctuation_string:
        example = example.replace(i, '')
    return example


def sep_lower(example):
    if example == '[SEP]' or example == '</s>' or example == '<s>':
        return example
    else:
        return example.lower()


class GraphFactory(object):

    def __init__(self, question_tokens, table_tokens, paragraph_tokens,
                 row_dict, column_dict, table_types, table, table_origin_types,
                 question_id, tokenizer):
        self.question_tokens = question_tokens
        self.table_tokens = table_tokens
        self.paragraph_tokens = paragraph_tokens
        self.global_tokens = question_tokens + table_tokens + paragraph_tokens
        self.table_types = table_types
        self.row_dict = row_dict
        self.column_dict = column_dict
        self.row_dict = OrderedDict(
            sorted(self.row_dict.items(), key=lambda x: x[0], reverse=False))
        self.column_dict = OrderedDict(
            sorted(self.column_dict.items(), key=lambda x: x[0],
                   reverse=False))
        self.graph = defaultdict(list)
        self.q_n = len(question_tokens)
        self.t_n = len(table_tokens)
        self.p_n = len(paragraph_tokens)
        self.table = table
        self.table_origin_types = table_origin_types
        self.question_id = question_id
        self.tokenizer = tokenizer

    def generate_graph(self):
        inner_table_relations = self.generate_inner_table_relation()
        inner_paragraph_relations = self.generate_inner_paragraph_relation()
        inner_question_relations = self.generate_inner_question_relation()
        question_table_relations = self.generate_question_table_relation()
        question_paragraph_relations = self.generate_question_paragraph_relation(
        )
        paragraph_table_relations = self.generate_paragraph_table_relation()
        global_relations = inner_table_relations+inner_paragraph_relations+\
            inner_question_relations+question_table_relations+question_paragraph_relations+\
                paragraph_table_relations
        global_relations = sorted(global_relations, key=lambda x: (x[0], x[1]))

        global_relations_new = []
        punctuation_string = string.punctuation
        for relation in global_relations:
            string1 = self.global_tokens[relation[0]]
            string2 = self.global_tokens[relation[1]]

            if string1 in punctuation_string or string2 in punctuation_string:
                if not ((relation[0] > self.q_n+self.t_n \
                and relation[1] > self.q_n+self.t_n)
                or (relation[0] <= self.q_n \
                and relation[1] <= self.q_n)\
                or (relation[0] <= self.q_n+self.t_n and relation[0]>self.q_n  \
                and relation[1] <= self.q_n+self.t_n and relation[0]>self.q_n )
                    ):
                    continue

            if relation[0]>self.q_n+self.t_n+self.p_n or \
                relation[1]>self.q_n+self.t_n+self.p_n:
                continue
            global_relations_new.append(relation)
        self.relations = sorted(global_relations_new,
                                key=lambda x: (x[0], x[1]))

        target_set = set(list(range(0, self.q_n + self.t_n + self.p_n)))
        current_set = set()
        for relation in self.relations:
            current_set.add(relation[0])
        cha = list(target_set - current_set)
        if cha:
            for chas in cha:
                self.relations.append((chas, chas, "self-same-self"))
        relation_set = list(set(self.relations))
        if len(relation_set) != len(self.relations):
            import pdb
            pdb.set_trace()

        graph = {}
        graph['src'] = list(map(lambda x: x[0], self.relations))
        graph['dst'] = list(map(lambda x: x[1], self.relations))
        graph['relations'] = list(
            map(lambda x: Relation_vocabulary['word2id'][x[2]],
                self.relations))
        graph['triplet'] = self.relations
        graph['dgl'] = dgl.graph((graph['src'], graph['dst']))

        return graph

    def type_check(self, idx):
        return self.table_types[idx]

    # 日期 1， row_head 3  column_head 4, number 2
    def generate_inner_table_relation(self):
        self.inner_table_relation = []

        for Row, R_list in self.row_dict.items():
            for i in range(len(R_list)):
                for j in range(len(R_list)):
                    if i == j:
                        continue
                    if self.type_check(R_list[i]) == 3 and self.type_check(R_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n + R_list[i], self.q_n + R_list[j],"Row-Contain-Cell"))
                    elif self.type_check(R_list[i]) == 2 and self.type_check(R_list[j]) == 3:
                        self.inner_table_relation.append((self.q_n + R_list[i], self.q_n + R_list[j],"Cell-Belong-Row"))
                    elif self.type_check(R_list[i]) == 1 and self.type_check(R_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n + R_list[i], self.q_n + R_list[j],"Time-Contain-Cell"))
                    elif self.type_check(R_list[i]) == 2 and self.type_check(R_list[j]) == 1:
                        self.inner_table_relation.append((self.q_n + R_list[i], self.q_n + R_list[j],"Cell-Belong-Time"))
                    elif self.type_check(R_list[i]) == 2 and self.type_check(R_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n + R_list[i], self.q_n + R_list[j],"Cell-SameRow-Cell"))
        for Column, C_list in self.column_dict.items():
            for i in range(len(C_list)):
                for j in range(len(C_list)):
                    if i == j:
                        continue
                    if self.type_check(C_list[i]) == 4 and self.type_check(C_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n + C_list[i], self.q_n + C_list[j],"Column-Contain-Cell"))
                    elif self.type_check(C_list[i]) == 2 and self.type_check(C_list[j]) == 4:
                        self.inner_table_relation.append((self.q_n + C_list[i], self.q_n + C_list[j],"Cell-Belong-Column"))
                    elif self.type_check(C_list[i]) == 1 and self.type_check(C_list[j]) == 2:
                        self.inner_table_relation.append((self.q_n + C_list[i], self.q_n + C_list[j],"Time-Contain-Cell"))
                    elif self.type_check(C_list[i]) == 2 and self.type_check(C_list[j]) == 1:
                        self.inner_table_relation.append((self.q_n + C_list[i], self.q_n + C_list[j],"Cell-Belong-Time"))

        xx, yy = self.table_origin_types.shape
        special_year = 0
        for i, row in enumerate(self.table_origin_types):
            if row[0] == 1 and np.sum(row) == 1 and i != 0:
                special_year += 1
        if special_year > 1:
            table_types_numpy = np.array(self.table_types)
            year_tags = np.where(table_types_numpy == 1)[0]
            for i in range(len(year_tags) - 1):
                start = year_tags[i]
                for j in range(year_tags[i] + 1, year_tags[i + 1]):
                    if self.table_types[j] == 2:
                        if (self.q_n + start, self.q_n + j, "Time-Contain-Cell") not in self.inner_table_relation:
                            self.inner_table_relation.append((self.q_n + start, self.q_n + j,"Time-Contain-Cell"))
                        if (self.q_n + j, self.q_n + start, "Cell-Belong-Time") not in self.inner_table_relation:
                            self.inner_table_relation.append((self.q_n + j, self.q_n + start,"Cell-Belong-Time"))
            start = year_tags[-1]
            for j in range(year_tags[-1] + 1, len(self.table_types)):
                if self.table_types[j] == 2:
                    if (self.q_n + start, self.q_n + j, "Time-Contain-Cell") not in self.inner_table_relation:
                        self.inner_table_relation.append((self.q_n + start, self.q_n + j,"Time-Contain-Cell"))
                    if (self.q_n + j, self.q_n + start, "Cell-Belong-Time") not in self.inner_table_relation:
                        self.inner_table_relation.append((self.q_n + j, self.q_n + start,"Cell-Belong-Time"))

        self.inner_table_relation = sorted(self.inner_table_relation,key=lambda x: (x[0], x[1]))

        return self.inner_table_relation

    def generate_inner_paragraph_relation(self):
        self.inner_paragraph_relation = []
        select_node = []
        for idx, node in enumerate(self.paragraph_tokens):
            if node == self.tokenizer.sep_token:
                select_node.append(idx)
        paragraph_split = []
        last_select_node = 0
        for idx, node in enumerate(self.paragraph_tokens):
            if node == self.tokenizer.sep_token:
                last_select_node = idx
            if node != self.tokenizer.sep_token:
                self.inner_paragraph_relation.append((self.q_n + self.t_n + last_select_node,self.q_n + self.t_n + idx, "Sentence-Contain-P_Word"))
                self.inner_paragraph_relation.append((self.q_n + self.t_n + idx,self.q_n + self.t_n + last_select_node,"P_Word-Partof-Sentence"))
        for i in range(len(self.paragraph_tokens) - 1):
            if self.paragraph_tokens[i] != self.tokenizer.sep_token and self.paragraph_tokens[i + 1] != self.tokenizer.sep_token:
                self.inner_paragraph_relation.append((self.q_n + self.t_n + i, self.q_n + self.t_n + i + 1,"Q_Word-PriorWord-Q_Word"))
                self.inner_paragraph_relation.append((self.q_n + self.t_n + i + 1, self.q_n + self.t_n + i,"Q_Word-NextWord-Q_Word"))
        self.inner_paragraph_relation = sorted(self.inner_paragraph_relation,key=lambda x: (x[0], x[1]))
        return self.inner_paragraph_relation

    def generate_inner_question_relation(self):
        self.inner_question_relations = []
        for i in range(len(self.question_tokens) - 1):
            self.inner_question_relations.append((i, i + 1, "Q_Word-PriorWord-Q_Word"))
            self.inner_question_relations.append((i + 1, i, "Q_Word-NextWord-Q_Word"))
        self.inner_question_relations = sorted(self.inner_question_relations, key=lambda x: (x[0], x[1]))
        return self.inner_question_relations

    def generate_question_table_relation(self):
        question = " ".join(self.question_tokens)
        question = question.lower()
        self.question_table_relations = []

        question_tokens = list(map(delete_punc, self.question_tokens))
        question_tokens = list(map(sep_lower, question_tokens))
        table_tokens = list(map(delete_punc, self.table_tokens))
        table_tokens = list(map(sep_lower, table_tokens))

        for i, word in enumerate(question_tokens):
            word = sep_lower(word)
            if word in stop_words or word == '':
                continue
            for j, cell in enumerate(table_tokens):
                flag1, flag2 = 0, 0
                cell = cell.strip(':')
                cell = sep_lower(cell)
                cell_words = cell.split(' ')
                if self.table_types[j] == 2:
                    continue
                if word in cell_words:
                    flag1 = 1
                if cell in question:
                    flag2 = 1
                if flag1 and flag2:
                    self.question_table_relations.append((i, self.q_n + j, "Q_Word-ExactMatch-Row"))
                    self.question_table_relations.append((self.q_n + j, i, "Row-ExactMatch-Q_Word"))
                elif flag1 and not flag2:
                    self.question_table_relations.append((i, self.q_n + j, "Q_Word-PartialMatch-Row"))
                    self.question_table_relations.append((j + self.q_n, i, "Row-PartialMatch-Q_Word"))
        self.question_table_relations = sorted(self.question_table_relations,key=lambda x: (x[0], x[1]))

        return self.question_table_relations

    def generate_question_paragraph_relation(self):
        self.question_paragraph_relations = []
        question_tokens = list(map(delete_punc, self.question_tokens))
        question_tokens = list(map(sep_lower, question_tokens))
        paragraph_tokens = list(map(delete_punc, self.paragraph_tokens))
        paragraph_tokens = list(map(sep_lower, paragraph_tokens))

        select_node = []
        for idx, node in enumerate(paragraph_tokens):
            if node == self.tokenizer.sep_token:
                select_node.append(idx)
        split_number = len(select_node)

        paragraph_split = [
            list(g) for k, g in groupby(
                paragraph_tokens, lambda x: x == self.tokenizer.sep_token)
            if not k
        ]

        for i, paragraph in enumerate(paragraph_split):
            for j, word in enumerate(question_tokens):
                if word in stop_words or word == '':
                    continue
                if word in paragraph:
                    try:
                        self.question_paragraph_relations.append((j, select_node[i] + self.q_n + self.t_n,"Q_Word-Partof-Sentence"))
                        self.question_paragraph_relations.append((select_node[i] + self.q_n + self.t_n, j,"Sentence-Contain-Q_Word"))
                    except:
                        import pdb
                        pdb.set_trace()

        for i, p_word in enumerate(paragraph_tokens):
            if p_word in stop_words or p_word == '':
                continue
            for j, word in enumerate(question_tokens):
                if p_word == word:
                    self.question_paragraph_relations.append((j, i + self.q_n + self.t_n, "Word-Same-Word"))
                    self.question_paragraph_relations.append((i + self.q_n + self.t_n, j, "Word-Same-Word"))
        self.question_paragraph_relations = sorted(self.question_paragraph_relations, key=lambda x: (x[0], x[1]))
        return self.question_paragraph_relations

    def generate_paragraph_table_relation(self):
        self.paragraph_table_relations = []
        table_tokens = list(map(delete_punc, self.table_tokens))
        table_tokens = list(map(sep_lower, table_tokens))
        paragraph_tokens = list(map(delete_punc, self.paragraph_tokens))
        paragraph_tokens = list(map(sep_lower, paragraph_tokens))

        select_node = []
        for idx, node in enumerate(paragraph_tokens):
            if node == self.tokenizer.sep_token:
                select_node.append(idx)
        split_number = len(select_node)

        paragraph_split = [
            list(g) for k, g in groupby(
                paragraph_tokens, lambda x: x == self.tokenizer.sep_token)
            if not k
        ]

        for id, paragraph_tokens_new in enumerate(paragraph_split):
            paragraph_tokens_new_new = paragraph_tokens_new.copy()
            a = paragraph_tokens_new_new.count('')
            for _ in range(a):
                paragraph_tokens_new_new.remove('')
            paragraph = " ".join(paragraph_tokens_new_new)
            for i, word in enumerate(paragraph_tokens_new):
                word = sep_lower(word)
                if word in stop_words or word == '':
                    continue
                for j, cell in enumerate(table_tokens):
                    flag1, flag2 = 0, 0
                    cell = cell.strip(':')
                    cell = sep_lower(cell)
                    cell_words = cell.split(' ')
                    if self.table_types[j] == 2:
                        continue
                    if word in cell_words:
                        flag1 = 1
                    if cell in paragraph:
                        flag2 = 1
                    if flag1 and flag2:
                        self.paragraph_table_relations.append((select_node[id] + i + 1 + self.q_n + self.t_n,j + self.q_n, "P_Word-ExactMatch-Row"))
                        self.paragraph_table_relations.append((j + self.q_n,select_node[id] + i + 1 + self.q_n + self.t_n,"Row-ExactMatch-P_Word"))
                    elif flag1 and not flag2:
                        self.paragraph_table_relations.append((select_node[id] + i + 1 + self.q_n + self.t_n,j + self.q_n, "P_Word-PartialMatch-Row"))
                        self.paragraph_table_relations.append((j + self.q_n,select_node[id] + i + 1 + self.q_n + self.t_n, "Row-PartialMatch-P_Word"))
            for k, cell2 in enumerate(table_tokens):
                cell2 = cell2.strip(':')
                cell2 = sep_lower(cell2)
                cell_words = cell2.split(' ')
                if self.table_types[k] == 2:
                    continue
                if cell2 in paragraph:
                    self.paragraph_table_relations.append((select_node[id] + self.q_n + self.t_n, k + self.q_n,"Sentence-Contain-Row"))
                    self.paragraph_table_relations.append((k + self.q_n, select_node[id] + self.q_n + self.t_n,"Row-Partof-Sentence"))
        self.paragraph_table_relations = sorted(self.paragraph_table_relations,key=lambda x: (x[0], x[1]))
        return self.paragraph_table_relations


class TaTQAReader(object):

    def __init__(self,
                 tokenizer,
                 passage_length_limit: int = None,
                 question_length_limit: int = None,
                 mode: str = 'train',
                 sep="<s>"):
        self.max_pieces = 512
        self.tokenizer = tokenizer
        self.passage_length_limit = passage_length_limit
        self.question_length_limit = question_length_limit
        # self.sep = self.tokenizer._convert_token_to_id(sep)
        self.sep = self.tokenizer.convert_tokens_to_ids(sep)
        self.skip_count = 0
        self.mode = mode

    def _make_instance(self, input_ids, attention_mask, token_type_ids, tags,
                       operator_class, question_tokens, table,
                       table_cell_type_ids, table_is_scale_type, table_number,
                       table_number_index, table_cell_index, table_tokens,
                       paragraph_tokens, paragraph_index, paragraph_number,
                       paragraph_number_index, answer_dict, question_id,
                       word_mask, number_mask, subword_lens, item_length,
                       word_word_mask, number_word_mask, table_word_tokens,
                       scale_class, table_word_tokens_number,
                       table_number_scale, paragraph_number_scale, output,
                       graph, question, paragraph_sep_tag, p_sep_fuzhu_tag,
                       answer_tp, question_number, question_number_index):
        return {
            "input_ids": np.array(input_ids),
            "attention_mask": np.array(attention_mask),
            "token_type_ids": np.array(token_type_ids),
            "tags": np.array(tags),
            "operator_class": int(operator_class),
            "question_tokens": question_tokens,
            "table": table,
            "table_cell_type_ids": table_cell_type_ids,
            "table_is_scale_type": table_is_scale_type,
            "table_number": table_number,
            "table_number_index": np.array(table_number_index),
            "table_cell_index": np.array(table_cell_index),
            "table_tokens": table_tokens,
            "paragraph_tokens": paragraph_tokens,
            "paragraph_index": np.array(paragraph_index),
            "paragraph_number": paragraph_number,
            "paragraph_number_index": np.array(paragraph_number_index),
            "answer_dict": answer_dict,
            "question_id": question_id,
            "word_mask": np.array(word_mask),
            "number_mask": np.array(number_mask),
            "subword_lens": np.array(subword_lens),
            "item_length": item_length,
            "word_word_mask": np.array(word_word_mask),
            "number_word_mask": np.array(number_word_mask),
            "table_word_tokens": table_word_tokens,
            "scale_class": scale_class,
            "table_word_tokens_number": table_word_tokens_number,
            "table_number_scale": table_number_scale,
            "paragraph_number_scale": paragraph_number_scale,
            "outputs": output,
            "graph": graph,
            "question": question,
            "paragraph_sep_tag": paragraph_sep_tag,
            "p_sep_fuzhu_tag": p_sep_fuzhu_tag,
            "answer_tp": answer_tp,
            "question_number": question_number,
            "question_number_index": question_number_index
        }

    def _to_instance(self, question: str, table: List[List[str]],
                     paragraphs: List[Dict], answer_from: str,
                     answer_type: str, answer: str, derivation: str,
                     facts: list, answer_mapping: Dict, scale: str,
                     question_id: str):

        table_pd,table_cell_type_ids,table_is_scale_type,table_cell_tokens,\
        table_cell_number_value,table_ids,table_tags,\
        table_cell_index, table_subword_lens, table_types, table_word_types,\
        table_word_tokens, table_word_tokens_number, table_cell_number_value_position,\
        row_dict, column_dict, table_cell_types_flatten = table_tokenize(table, self.tokenizer, answer_mapping,question_id)

        paragraph_tokens, paragraph_ids, paragraph_tags, number_value, \
        paragraph_index, paragraph_subword_lens,paragraph_types, paragraph_word_types, paragraph_number_scale, paragraph_sep_tag\
        = paragraph_tokenize(question, paragraphs, self.tokenizer, answer_mapping, question_id)

        question_ids, question_tokens, question_subword_lens, question_word_index, question_types, question_word_types, question_number_value, question_tags\
        = question_tokenize(question, self.tokenizer, answer_mapping)

        table_cell_number_scale = generate_number_scale(
            table, table_pd, table_cell_type_ids, table_is_scale_type,
            table_cell_number_value, table_cell_number_value_position,
            paragraphs)

        input_ids, attention_mask, input_types, input_word_types, subword_lens, tags,\
        table_cell_index,table_tokens,table_number, table_number_index, paragraph_index,\
        paragraph_tokens, paragraph_number, paragraph_number_index,token_type_ids, \
        table_number_position,table_number_scale, paragraph_number_scale, question_number, question_number_index,\
        row_dict, column_dict, table_cell_types_flatten, paragraph_sep_tag, p_sep_fuzhu_tag, answer_tp \
        = _concat(question_ids, table_ids, paragraph_ids, table_tags, paragraph_tags,\
        question_tokens,table_cell_tokens, paragraph_tokens,table_cell_number_value,number_value,\
        table_cell_index, paragraph_index,question_subword_lens,question_types,table_subword_lens,\
        table_types, paragraph_subword_lens,paragraph_types,question_word_types,table_word_types,\
        paragraph_word_types,table_cell_number_value_position,table_cell_number_scale, paragraph_number_scale, question_number_value, question_tags,\
        row_dict, column_dict, table_cell_types_flatten, paragraph_sep_tag,\
        self.sep,self.question_length_limit,self.passage_length_limit, self.max_pieces,question_id,self.tokenizer)
        output = []

        if self.mode == 'train':
            output = generate_output2(answer_type, derivation, question_number, table_number, paragraph_number, question, table, paragraphs, answer_mapping, facts, table_number_position, answer, table_pd, question_id, paragraph_tokens)

        answer_dict = {
            "answer_type": answer_type,
            "answer": answer,
            "scale": scale,
            "answer_from": answer_from
        }
        operator_class = generate_class(answer_dict, answer_mapping)
        try:
            scale_class = SCALE.index(scale)
        except:
            import pdb
            pdb.set_trace()
        item_length = {
            'q': len(question_tokens),
            't': len(table_tokens),
            'p': len(paragraph_tokens)
        }

        word_mask = torch.zeros(512)
        number_mask = torch.zeros(512)

        try:
            for i, idx in enumerate(input_types):
                if idx == 3:
                    word_mask[i] = 1
                if idx == 2:
                    number_mask[i] = 1
        except:
            import pdb; 
            pdb.set_trace()

        word_word_mask = []
        number_word_mask = []
        for i, idx in enumerate(input_word_types):
            if idx == 3:
                word_word_mask += [1]
                number_word_mask += [0]
            if idx == 2:
                word_word_mask += [0]
                number_word_mask += [1]

        graphFactory = GraphFactory(question_tokens, table_tokens,
                                    paragraph_tokens, row_dict, column_dict,
                                    table_cell_types_flatten, table_pd,
                                    table_cell_type_ids, question_id,
                                    self.tokenizer)
        graph = graphFactory.generate_graph()

        return self._make_instance(
            input_ids, attention_mask, token_type_ids, tags, operator_class,
            question_tokens, table_pd, table_cell_type_ids,
            table_is_scale_type, table_number, table_number_index,
            table_cell_index, table_tokens, paragraph_tokens, paragraph_index,
            paragraph_number, paragraph_number_index, answer_dict, question_id,
            word_mask, number_mask, subword_lens, item_length, word_word_mask,
            number_word_mask, table_word_tokens, scale_class,
            table_word_tokens_number, table_number_scale,
            paragraph_number_scale, output, graph, question, paragraph_sep_tag,
            p_sep_fuzhu_tag, answer_tp, question_number, question_number_index)

    def _read(self, file_path: str):
        print("Reading file at %s", file_path)
        with open(file_path) as dataset_file:
            dataset = json.load(dataset_file)
        instances = []
        random.shuffle(dataset)
        count = 0
        total = 0
        total_count = defaultdict(int)
        for one in tqdm(dataset):
            table = one['table']['table']
            paragraphs = one['paragraphs']
            questions = one['questions']

            for question_answer in questions:
                question = question_answer["question"].strip()
                answer_type = question_answer["answer_type"]
                derivation = question_answer["derivation"]
                answer = question_answer["answer"]
                answer_mapping = question_answer["mapping"]
                facts = question_answer["facts"]
                answer_from = question_answer["answer_from"]
                scale = question_answer["scale"]
                question_id = question_answer["uid"]
                instance = self._to_instance(question, table, paragraphs,
                                             answer_from, answer_type, answer,
                                             derivation, facts, answer_mapping,
                                             scale, question_id)
                if answer_type == 'arithmetic':
                    count += 1
                    total += len(instance['outputs'])
                    total_count[len(instance['outputs'])] += 1
                if instance is not None:
                    instances.append(instance)
        print(len(instances))
        print(total_count)
        print(total / count)
        return instances



"""
Use this tool for different models and datasets to solve table related tasks.
"""
import pandas as pd
import logging
import os
import torch
import sys
from functools import partial
from typing import List, Optional
from abc import ABC, abstractmethod
import nltk  # Here to have a nice missing dependency error message early on
import numpy as np
import json
from datasets import load_dataset
import transformers
from filelock import FileLock
from transformers import (
    AutoConfig,
    BartForConditionalGeneration,
    DataCollatorForSeq2Seq,
    HfArgumentParser,
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments,
    TapexTokenizer,
    AutoTokenizer,
    TapasTokenizer,
    TapasForQuestionAnswering,
    set_seed,
)
from transformers.file_utils import is_offline_mode
from transformers.trainer_utils import get_last_checkpoint, is_main_process
from transformers.utils import check_min_version
from utils.dataset import BaseStructuredQADataset
from utils.common import *

check_min_version("4.17.0.dev0")

logger = logging.getLogger(__name__)
os.environ["WANDB_PROJECT"] = "reastap_march" # .

try:
    nltk.data.find("tokenizers/punkt")
except (LookupError, OSError):
    if is_offline_mode():
        raise LookupError(
            "Offline mode: run this script without TRANSFORMERS_OFFLINE first to download nltk data files"
        )
    with FileLock(".lock") as lock:
        nltk.download("punkt", quiet=True)


class ToolBaseForStructuredQA(ABC):

    @abstractmethod
    def __init__(self, model_name_or_path, **kwargs) -> None:
        r"""
        对模型进行初始化, 加载模型权重, 
        固定 model 和 tokenizer

        Params:
            model_name_or_path (`str` or `os.PathLike`):
                Path or name to your model. Default: `Yale-LILY/reastap-large`
            tokenizer_name (`str` , *optional*):
                Name of tokenizer to be loaded. Default to model_name_or_path.
        """
        args_dict = DEFAULT_MODEL_ARGUMENTS.copy()
        args_dict['model_name_or_path'] = model_name_or_path
        args_dict.update(kwargs)
        parser = HfArgumentParser(ModelArguments)
        model_args,  = parser.parse_dict(args_dict)
        self.model_args = model_args
        self.dataset = None
        self.training_args = None
        # Setup logging
        logging.basicConfig(
            format="%(asctime)s - %(levelname)s - %(name)s -   %(message)s",
            datefmt="%m/%d/%Y %H:%M:%S",
            handlers=[logging.StreamHandler(sys.stdout)],
        )
        # Load pretrained model and tokenizer
        self._load_model(model_args)

    @abstractmethod
    def _load_model(self, model_args):
        print(f'Load Model: {model_args.model_name_or_path}')

    @abstractmethod
    def _init_running_cfg(self, default_cfg, **kwargs):
        pass
    
    # def run_wikisql(self):
    #     # TODO: 
    #     pass

    # def run_wtq(self):
    #     # TODO:
    #     pass

    # def run(self):
    #     # TODO:
    #     pass

    def train(self, 
              dataset: Optional[BaseStructuredQADataset] = None, 
              **kwargs):

        training_args = self._init_running_cfg(DEFAULT_TRAINING_ARGUMENTS,**kwargs)
        if dataset == None:
            raise ValueError("requires a dataset, like wikisql, wtq, e.g.")
        if dataset.train_dataset == None:
            raise ValueError("requires a train dataset")
        if dataset.eval_dataset == None:
            raise ValueError("requires a validation dataset")

        preprocess_fn = partial(self.preprocess_tableqa_function, sample_fn = dataset.get_example)
        train_dataset = dataset.train_dataset
        column_names = train_dataset.column_names

        if training_args.max_train_samples is not None:
            train_dataset = train_dataset.select(range(training_args.max_train_samples))
        preprocess_fn_training = partial(preprocess_fn, is_training=True)
        train_dataset = train_dataset.map(
            preprocess_fn_training,
            batched=True,
            num_proc=training_args.preprocessing_num_workers,
            remove_columns=column_names,
            load_from_cache_file=not training_args.overwrite_cache,
        )
        eval_dataset = dataset.eval_dataset
        column_names = eval_dataset.column_names

        if training_args.max_eval_samples is not None:
            eval_dataset = eval_dataset.select(range(training_args.max_eval_samples))
        preprocess_fn_eval = preprocess_fn
        eval_dataset = eval_dataset.map(
            preprocess_fn_eval,
            batched=True,
            num_proc=training_args.preprocessing_num_workers,
            remove_columns=column_names,
            load_from_cache_file=not training_args.overwrite_cache,
        )
        self.last_checkpoint = None
        if os.path.isdir(training_args.output_dir) and training_args.do_train and not training_args.overwrite_output_dir:
            self.last_checkpoint = get_last_checkpoint(training_args.output_dir)
            if self.last_checkpoint is None and len(os.listdir(training_args.output_dir)) > 0:
                raise ValueError(
                    f"Output directory ({training_args.output_dir}) already exists and is not empty. "
                    "Use --overwrite_output_dir to overcome."
                )
            elif self.last_checkpoint is not None and training_args.resume_from_checkpoint is None:
                logger.info(
                    f"Checkpoint detected, resuming training at {self.last_checkpoint}. To avoid this behavior, change "
                    "the `--output_dir` or add `--overwrite_output_dir` to train from scratch."
                )
        trainer = Seq2SeqTrainer(
            model=self.model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset= eval_dataset,
            tokenizer=self.tokenizer,
            data_collator=self.data_collator,
            compute_metrics=self.compute_metrics
        )

        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        elif self.last_checkpoint is not None:
            checkpoint = self.last_checkpoint
        train_result = trainer.train(resume_from_checkpoint=checkpoint)
        trainer.save_model()  # Saves the tokenizer too for easy upload

        metrics = train_result.metrics
        max_train_samples = (
            training_args.max_train_samples if training_args.max_train_samples is not None else len(train_dataset)
        )
        metrics["train_samples"] = min(max_train_samples, len(train_dataset))

        trainer.log_metrics("train", metrics)
        trainer.save_metrics("train", metrics)
        trainer.save_state()

        return metrics
        
    def eval(self, 
              dataset: Optional[BaseStructuredQADataset] = None, 
              **kwargs):
        training_args = self._init_running_cfg(DEFAULT_EVAL_ARGUMENTS,**kwargs)
        if dataset == None:
            raise ValueError("requires a dataset, like wikisql, wtq, e.g.")
        
        preprocess_fn = partial(self.preprocess_tableqa_function, sample_fn = dataset.get_example)
        eval_dataset = dataset.eval_dataset
        column_names = eval_dataset.column_names
        
        if training_args.max_eval_samples is not None:
            eval_dataset = eval_dataset.select(range(training_args.max_eval_samples))

        preprocess_fn_eval = preprocess_fn


        ########################
        ## inject: test tapas tokenizer
        # BEGIN = 4872
        # BATCH_SIZE = 12
        # for i in range(BEGIN,BEGIN+BATCH_SIZE): ## [1480, 4881] 挂了
        #     print(i)
        #     row_data = eval_dataset[i]

        #     sample_input = dict()

        #     for key, value in row_data.items():
        #         # print(key)
        #         sample_input[key] = [value]
            
        #     input = preprocess_fn_eval(examples=sample_input)

        #     model = self.model
        #     model_input = input
        #     # model_input = dict()

        #     # for key, value in input.items():
        #     #     import pdb; pdb.set_trace()
        #     #     if key in ['input_ids', 'labels', 'token_type_ids', 'attention_mask']:
        #     #         model_input[key] = torch.tensor(value, dtype=torch.int64)
        #     #     else:
        #     #         model_input[key] = torch.tensor(value)
        #     for key, value in input.items():
        #         if value.max() == 151:
        #             import pdb; pdb.set_trace()
            
        #     if len(input)==0:
        #         continue
        #     try:
        #         outputs = model(**model_input)
        #     except Exception as e:
        #         print(e)
        #         import pdb; pdb.set_trace()

        # import pdb; pdb.set_trace()
        #########################

        eval_dataset = eval_dataset.map(
            preprocess_fn_eval,
            batched=True,
            num_proc=training_args.preprocessing_num_workers,
            remove_columns=column_names,
            load_from_cache_file=not training_args.overwrite_cache,
        )

        ########################
        ## inject: test tapas model
        # for i in range(980,983):
        #     input = eval_dataset[i]
        #     model = self.model
        #     model_input = dict()
        #     for key, value in input.items():
        #         if key in ['input_ids', 'labels', 'token_type_ids', 'attention_mask']:
        #             model_input[key] = torch.tensor([value], dtype=torch.int64)
        #         else:
        #             model_input[key] = torch.tensor([value])
        #     import pdb; pdb.set_trace()
        #     outputs = model(**model_input)
        #     loss = outputs.loss
        #     pdb.set_trace()
        #########################

        # Initialize Trainer
        trainer = Seq2SeqTrainer(
            model=self.model,
            args=training_args,
            train_dataset= None,
            eval_dataset=eval_dataset,
            tokenizer=self.tokenizer,
            data_collator=self.data_collator,
            compute_metrics=self.compute_metrics
        )

        # Evaluation
        # results = {}
        logger.info("*** Evaluate ***")

        metrics = trainer.evaluate(
            max_length=training_args.val_max_target_length, num_beams=training_args.num_beams, metric_key_prefix="eval"
        )
        max_eval_samples = training_args.max_eval_samples if training_args.max_eval_samples is not None else len(eval_dataset)
        metrics["eval_samples"] = min(max_eval_samples, len(eval_dataset))

        trainer.log_metrics("eval", metrics)

        return metrics

    def predict(self, dataset=None, **kwargs):
        training_args = self._init_running_cfg(DEFAULT_PREDICT_ARGUMENTS, **kwargs)
        max_target_length = training_args.val_max_target_length

        if dataset == None:
            raise ValueError("requires a dataset, like wikisql, wtq, e.g.")

        preprocess_fn = partial(self.preprocess_tableqa_function, sample_fn = dataset.get_example)
        test_dataset = dataset.test_dataset
        raw_test_dataset = test_dataset
        column_names = test_dataset.column_names
        if training_args.max_predict_samples is not None:
            test_dataset = test_dataset.select(range(training_args.max_predict_samples))
        preprocess_fn_test = preprocess_fn
        test_dataset = test_dataset.map(
            preprocess_fn_test,
            batched=True,
            num_proc=training_args.preprocessing_num_workers,
            remove_columns=column_names,
            load_from_cache_file=not training_args.overwrite_cache,
        )
        
        # Initialize Trainer
        trainer = Seq2SeqTrainer(
            model=self.model,
            args=training_args,
            train_dataset= None,
            eval_dataset= None,
            tokenizer=self.tokenizer,
            data_collator=self.data_collator,
            compute_metrics=self.compute_metrics
        )

        logger.info("*** Predict ***")

        predict_results = trainer.predict(
            test_dataset,
            metric_key_prefix="predict",
            max_length=training_args.val_max_target_length,
            num_beams=training_args.num_beams,
        )
        metrics = predict_results.metrics
        max_predict_samples = (
            training_args.max_predict_samples if training_args.max_predict_samples is not None else len(test_dataset)
        )
        metrics["predict_samples"] = min(max_predict_samples, len(test_dataset))

        trainer.log_metrics("predict", metrics)
        
        predictions = self.tokenizer.batch_decode(
            predict_results.predictions, skip_special_tokens=True, clean_up_tokenization_spaces=True
        )
        predictions = [pred.strip() for pred in predictions]

        outputs = []

        # NOTE: 这里的答案应该只针对 wikisql 和 wtq 数据集, 别的应该没有
        # for i, pred in enumerate(predictions):
        #     example = raw_test_dataset[i]
        #     if self.seek_sql:
        #         table = convert_table_types(example["table"])
        #         answer = retrieve_wikisql_query_answer_tapas(table, example["sql"])

        #         answer = ", ".join([i.lower() for i in answer])
        #         example_id = str(i)
        #     else:
        #         answer = ", ".join([i.lower() for i in example["answers"]])
        #         example_id = example["id"]

        #     outputs.append({"id": example_id, "prediction": pred, "answers": answer})

        for i, pred in enumerate(predictions):
            example = raw_test_dataset[i]
            if 'id' in example.keys():
                outputs.append({"id": example["id"], "prediction": pred})
            else :
                outputs.append({"id": str(i), "prediction": pred}) 

        model_name = self.model_args.model_name_or_path.split("/")[-1].split("-")[0]
        if training_args.output_dir:
            output_prediction_file = os.path.join(training_args.output_dir, f"{model_name}_{dataset.name}_{training_args.inference_set}_predictions.json")
            json.dump(outputs, open(output_prediction_file, "w"), indent=4)
        
        return outputs


    @abstractmethod
    def preprocess_tableqa_function(self, examples, sample_fn = None, is_training=False):
        pass
    

class BertForStructuredQA(ToolBaseForStructuredQA):
    """
    Bert-based TableQA Tools (TAPAS)
    """
    def __init__(self, model_name_or_path, **kwargs) -> None:
        super().__init__(model_name_or_path, **kwargs)


    def _load_model(self, model_args):
        super()._load_model(model_args)
        config = AutoConfig.from_pretrained(
            model_args.config_name if model_args.config_name else model_args.model_name_or_path,
            cache_dir=model_args.cache_dir,
            revision=model_args.model_revision,
            use_auth_token=True if model_args.use_auth_token else None,
        )
        # NOTE: we do not use `aggregation_labels` now
        config.num_aggregation_labels = 0

        self.tokenizer = TapasTokenizer.from_pretrained(
            model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
            cache_dir=model_args.cache_dir,
            use_fast=model_args.use_fast_tokenizer,
            revision=model_args.model_revision,
            use_auth_token=True if model_args.use_auth_token else None,
            add_prefix_space=True,
        )

        self.model = TapasForQuestionAnswering.from_pretrained(
            model_args.model_name_or_path,
            from_tf=bool(".ckpt" in model_args.model_name_or_path),
            config=config,
            cache_dir=model_args.cache_dir,
            revision=model_args.model_revision,
            use_auth_token=True if model_args.use_auth_token else None,
        )

    def _init_running_cfg(self, default_cfg, **kwargs):
        """
        initialize running config and data collator.
        special for bert-based model.
        """
        args_dict = default_cfg.copy()
        args_dict["max_source_length"] = 512
        args_dict["max_target_length"] = None
        args_dict["predict_with_generate"] = False
        args_dict["pad_to_max_length"]= True
        args_dict["val_max_target_length"] = None
        args_dict["label_names"] = ['labels'] # NOTE: default to use only `labels` as labels
        args_dict.update(kwargs)

        parser = HfArgumentParser(TTQA_TrainingArguments)

        training_args, = parser.parse_dict(args_dict)
        self.training_args = training_args
        
        logger.setLevel(logging.INFO if is_main_process(training_args.local_rank) else logging.WARN)

        # Log on each process the small summary:
        logger.warning(
            f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
            + f"distributed training: {bool(training_args.local_rank != -1)}, 16-bits training: {training_args.fp16}"
        )
        # Set the verbosity to info of the Transformers logger (on main process only):
        if is_main_process(training_args.local_rank):
            transformers.utils.logging.set_verbosity_info()
        logger.info(f"Training/evaluation parameters {training_args}")

        # Set seed before initializing model.
        set_seed(training_args.seed)
        # Temporarily set max_target_length for training.
        self.padding = "max_length" if training_args.pad_to_max_length else False

        if training_args.label_smoothing_factor > 0 and not hasattr(self.model, "prepare_decoder_input_ids_from_labels"):
            logger.warning(
                "label_smoothing is enabled but the `prepare_decoder_input_ids_from_labels` method is not defined for"
                f"`{self.model.__class__.__name__}`. This will lead to loss being calculated twice and will take up more memory"
            )
        
        # Data collator
        self.label_pad_token_id = -100 if training_args.ignore_pad_token_for_loss else self.tokenizer.pad_token_id
        self.data_collator = DataCollatorForSeq2Seq(
            self.tokenizer,
            model=self.model,
            label_pad_token_id=self.label_pad_token_id,
            pad_to_multiple_of=8 if training_args.fp16 else None,
        )

        compute_metrics_fn = args_dict['compute_metrics'] if 'compute_metrics' in args_dict else partial(compute_metrics_tapas, tokenizer = self.tokenizer)

        self.compute_metrics = compute_metrics_fn
        return training_args

    def preprocess_tableqa_function(self, examples, sample_fn = None, is_training=False):
        """
        The is_training FLAG is used to identify if we could use the supervision
        to truncate the table content if it is required.

        seek_sql FLAG is used to identify whether to retrieve the WikiSQL answer for each question
        """
        if sample_fn == None:
            raise ValueError("should have a sample_fn")

        questions, tables, answers, answer_coordinates = sample_fn(examples)
        # IMPORTANT: we cannot pass by answers during evaluation, answers passed during training are used to
        # truncate large tables in the train set!

        # NOTE: tapas tokenizer only support tokenizing one table one time
        model_inputs = dict()

        
        # if is_training: # NOTE: it seems useless in tapas
        if answers is not None and len(answers) > 0:
            for table, query, answer,answer_coordinates in zip(tables, questions, answers, answer_coordinates):
                tuple_coordinates_list = [tuple(item) for item in answer_coordinates]
                try:
                    tokenize_result = self.tokenizer(
                        table=table.astype(str),
                        queries=query,
                        answer_text=answer,
                        answer_coordinates = tuple_coordinates_list,
                        max_length=self.training_args.max_source_length,
                        padding=self.padding,
                        truncation=True,
                        return_tensors="pt",
                    )
                except:
                    print("WARNING: One table is too large to tokenize, remove it.")
                    continue

                if tokenize_result['token_type_ids'][:,:,4:6].max() > 127: # TODO: fix tokenizer
                    print("WARNING: There is a dirty data, remove it.")
                    continue

                for (key ,value) in tokenize_result.items():
                    if key not in model_inputs.keys():
                        model_inputs[key] = value
                    else:
                        model_inputs[key] = torch.cat((model_inputs[key],value),dim=0)
        else:
            for table, query in zip(tables, questions):
                tokenize_result = self.tokenizer(
                    table=table.astype(str),
                    queries=query,
                    max_length=self.training_args.max_source_length,
                    padding=self.padding,
                    truncation=True,
                    return_tensors="pt",
                )
                for (key ,value) in tokenize_result.items():
                    if key not in model_inputs.keys():
                        model_inputs[key] = value
                    else:
                        model_inputs[key] = torch.cat((model_inputs[key],value),dim=0)

        return model_inputs



class BartForStructuredQA(ToolBaseForStructuredQA):
    """
    Bart-based TextTableQA Tools
    """

    def __init__(self, model_name_or_path, **kwargs) -> None:
        super().__init__(model_name_or_path, **kwargs)
    
    def _load_model(self, model_args):
        super()._load_model(model_args)
        config = AutoConfig.from_pretrained(
            model_args.config_name if model_args.config_name else model_args.model_name_or_path,
            cache_dir=model_args.cache_dir,
            revision=model_args.model_revision,
            use_auth_token=True if model_args.use_auth_token else None,
        )

        config.max_length = 1024
        config.early_stopping = False

        # load tapex tokenizer
        self.tokenizer = TapexTokenizer.from_pretrained(
            model_args.tokenizer_name if model_args.tokenizer_name else model_args.model_name_or_path,
            cache_dir=model_args.cache_dir,
            use_fast=model_args.use_fast_tokenizer,
            revision=model_args.model_revision,
            use_auth_token=True if model_args.use_auth_token else None,
            add_prefix_space=True,
        )

        # load Bart based Tapex model (default tapex-large)
        self.model = BartForConditionalGeneration.from_pretrained(
            model_args.model_name_or_path,
            from_tf=bool(".ckpt" in model_args.model_name_or_path),
            config=config,
            cache_dir=model_args.cache_dir,
            revision=model_args.model_revision,
            use_auth_token=True if model_args.use_auth_token else None,
        )

        if self.model.config.decoder_start_token_id is None:
            raise ValueError("Make sure that `config.decoder_start_token_id` is correctly defined")

    def _init_running_cfg(self, default_cfg, **kwargs):
        args_dict = default_cfg.copy()
        args_dict.update(kwargs)
        parser = HfArgumentParser(TTQA_TrainingArguments)
        training_args, = parser.parse_dict(args_dict)
        self.training_args = training_args
        
        logger.setLevel(logging.INFO if is_main_process(training_args.local_rank) else logging.WARN)

        # Log on each process the small summary:
        logger.warning(
            f"Process rank: {training_args.local_rank}, device: {training_args.device}, n_gpu: {training_args.n_gpu}"
            + f"distributed training: {bool(training_args.local_rank != -1)}, 16-bits training: {training_args.fp16}"
        )
        # Set the verbosity to info of the Transformers logger (on main process only):
        if is_main_process(training_args.local_rank):
            transformers.utils.logging.set_verbosity_info()
        logger.info(f"Training/evaluation parameters {training_args}")

        # Set seed before initializing model.
        set_seed(training_args.seed)
        # Temporarily set max_target_length for training.
        self.max_target_length = training_args.max_target_length
        self.padding = "max_length" if training_args.pad_to_max_length else False

        if training_args.label_smoothing_factor > 0 and not hasattr(self.model, "prepare_decoder_input_ids_from_labels"):
            logger.warning(
                "label_smoothing is enabled but the `prepare_decoder_input_ids_from_labels` method is not defined for"
                f"`{self.model.__class__.__name__}`. This will lead to loss being calculated twice and will take up more memory"
            )
        
        # Data collator
        self.label_pad_token_id = -100 if training_args.ignore_pad_token_for_loss else self.tokenizer.pad_token_id
        self.data_collator = DataCollatorForSeq2Seq(
            self.tokenizer,
            model=self.model,
            label_pad_token_id=self.label_pad_token_id,
            pad_to_multiple_of=8 if training_args.fp16 else None,
        )
        compute_metrics_fn = args_dict['compute_metrics'] if 'compute_metrics' in args_dict else compute_metrics

        self.compute_metrics = partial(compute_metrics_fn, tokenizer = self.tokenizer) if training_args.predict_with_generate else None

        return training_args


    def preprocess_tableqa_function(self, examples, sample_fn = None, is_training=False):
        """
        The is_training FLAG is used to identify if we could use the supervision
        to truncate the table content if it is required.

        seek_sql FLAG is used to identify whether to retrieve the WikiSQL answer for each question
        """
        if sample_fn == None:
            raise ValueError("should have a sample_fn")
        
        questions, tables, answers = sample_fn(examples)[:3]

        # IMPORTANT: we cannot pass by answers during evaluation, answers passed during training are used to
        # truncate large tables in the train set!
        if is_training:
            model_inputs = self.tokenizer(
                table=tables,
                query=questions,
                answer=answers,
                max_length=self.training_args.max_source_length,
                padding=self.padding,
                truncation=True,
            )
        else:
            model_inputs = self.tokenizer(
                table=tables, query=questions, max_length=self.training_args.max_source_length, padding=self.padding, truncation=True
            )
            
        labels = self.tokenizer(
            answer=[", ".join(answer) for answer in answers],
            max_length=self.max_target_length,
            padding=self.padding,
            truncation=True,
        )

        # If we are padding here, replace all tokenizer.pad_token_id in the labels by -100 when we want to ignore
        # padding in the loss.
        if self.padding == "max_length" and self.training_args.ignore_pad_token_for_loss:
            labels["input_ids"] = [
                [(l if l != self.tokenizer.pad_token_id else -100) for l in label] for label in labels["input_ids"]
            ]

        model_inputs["labels"] = labels["input_ids"]
        return model_inputs

if __name__ == '__main__':
    tool = BertForStructuredQA(model_name_or_path = '/home/lfy/PTM/tapex-large-finetuned-sqa')


import torch
import os
import math
from transformers import get_linear_schedule_with_warmup, AdamW
import argparse
import random
import numpy as np
from torch import nn
from torch.nn import functional as F
import logging
from transformers import AutoTokenizer, HfArgumentParser
from tqdm import tqdm
from torch.utils.data import DataLoader
import sys
from time import gmtime, strftime
import json
from .dataset import ClassifyDataset,collate
from .model import ClassifierModel
from .utils import get_training_args

ROOT_DIR = os.path.join(os.path.dirname(__file__), "../")
MAX_LEN = 512
TYPES = ['image', 'text', 'table', 'compose']
LABELS = {
    "image": 0,
    "text": 1,
    "table": 2,
    "compose": 3
}

def get_type(type_str):
    if type_str in ['ImageQ', 'ImageListQ']:
        return 'image'
    elif type_str in ['TableQ']:
        return 'table'
    elif type_str in ['TextQ']:
        return 'text'
    else:
        return 'compose'
    
def get_keys_by_value(dict_obj, value):
    keys = []
    for k, v in dict_obj.items():
        if v == value:
            keys.append(k)
    return keys

def create_logger(name, silent=False, to_disk=True, log_file=None):
    """Logger wrapper
    """
    # setup logger
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.propagate = False
    formatter = logging.Formatter(fmt='%(asctime)s %(message)s', datefmt='%m/%d/%Y %I:%M:%S')
    if not silent:
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log.addHandler(ch)
    if to_disk:
        log_file = log_file if log_file is not None else strftime("%Y-%m-%d-%H-%M-%S.log", gmtime())
        fh = logging.FileHandler(log_file)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        log.addHandler(fh)
    return log

class ClassifierTrainer:
    def __init__(self, type_fn = get_type, type_map = LABELS) -> None:
        args = get_training_args()
        self.args = args
        if not os.path.exists(args.output_dir):
            os.mkdir(args.output_dir)
        self.model = ClassifierModel(args.bert_model, args.num_classes, args.dropout)
        # model.to(device)
        self.tokenizer = AutoTokenizer.from_pretrained(args.bert_model)
        self.train_set = None
        self.dev_set = None
        self.test_set = None
        self.type_fn = type_fn
        self.type_map = type_map
        self.loss_fn = nn.CrossEntropyLoss() if args.loss_fn_type == 'CE' else nn.BCEWithLogitsLoss(reduction='sum')
            

    def train_epoch(self, model, loader, optimizer, scheduler, logger):
        model.train()
        averge_step = len(loader) // 20
        loss_sum, step = 0, 0
        for i, data in enumerate(tqdm(loader)):
            probs = model(data) # probs for each type
            # if sum(data['labels'][0]).cpu().item()!=1:
                # data['labels'] = F.softmax(probs + (1 - data['labels'].float()) * -1e20, dim=-1)
            loss = self.loss_fn(probs, data['labels'])
            optimizer.zero_grad()
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 0.1)
            optimizer.step()
            scheduler.step()
            loss_sum += loss
            step += 1
            if i % averge_step == 0:
                logger.info("Training Loss [{0:.5f}]".format(loss_sum/step))
                loss_sum, step = 0, 0

    def train(self, resume_from = None):
        logger = create_logger("Training", log_file=os.path.join(self.args.output_dir, self.args.log_file))
        if self.train_set == None:
            self.train_set = ClassifyDataset(tokenizer=self.tokenizer,type_fn=self.type_fn, type_map=self.type_map,
                                             data_path=self.args.data_path, split='train')
        if self.dev_set == None:
            self.dev_set = ClassifyDataset(tokenizer=self.tokenizer,type_fn=self.type_fn, type_map=self.type_map,
                                           data_path=self.args.data_path, split='dev')

        train_loader = DataLoader(self.train_set, batch_size=self.args.per_device_train_batch_size, 
                                  collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length))
        
        dev_loader = DataLoader(self.dev_set, batch_size=self.args.per_device_eval_batch_size, 
                                collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length))
        
        epochs = self.args.num_train_epochs        
        device = torch.device("cuda")
        model = self.model
        model.to(device)
        if self.args.n_gpu > 1:
            model = torch.nn.DataParallel(model)

        optimizer = AdamW(model.parameters(), lr=self.args.learning_rate, eps=self.args.adam_epsilon)

        t_total = len(self.train_set) * epochs

        scheduler = get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=self.args.warmup_steps * t_total, num_training_steps=t_total
        )

        # outputs/classifier.pt 是否存在？
        # load model
        best_acc = 0
        model_load_path = self.args.resume_path if resume_from is None else resume_from
        if model_load_path is not None and os.path.exists(model_load_path):
            logger.info(f"loading trained parameters from {model_load_path}")
            model.load_state_dict(torch.load(model_load_path))
            logger.info("start eval....")
            acc = self.eval(dev_loader, logger)
            logger.info(f"acc: {acc}")
            best_acc = acc

        for epoch in range(epochs):
            logger.info(f"Training epoch: {epoch}")
            self.train_epoch(model, train_loader, optimizer, scheduler, logger)
            logger.info("start eval....")
            acc = self.eval(dev_loader, logger)
            logger.info(f"[{epoch}/{epochs}] acc: {acc}")
            if acc > best_acc:
                best_acc = acc
                model_to_save = model.module if hasattr(model, "module") else model
                model_save_path = os.path.join(self.args.output_dir, self.args.ckpt_file)
                logger.info(f"saving model...to {model_save_path}")
                torch.save(model_to_save.state_dict(), model_save_path)


    def eval(self, loader = None, logger=None, save = False, ckpt_path = None):
        if logger == None:
            logger = create_logger("Evaling", log_file=os.path.join(self.args.output_dir, self.args.log_file))


        if loader == None:
            if self.dev_set == None:
                self.dev_set = ClassifyDataset(tokenizer=self.tokenizer,type_fn=self.type_fn, type_map=self.type_map,
                                               data_path=self.args.data_path, split='dev')
            loader = DataLoader(self.dev_set, batch_size=self.args.per_device_eval_batch_size, 
                                    collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length))

        model = self.model
        if ckpt_path != None:
            """eval independently"""
            model_load_path = ckpt_path
            logger.info(f"loading trained parameters from {model_load_path}")
            model.load_state_dict(torch.load(model_load_path))
            logger.info("start eval....")
            device = torch.device("cuda")
            model.to(device)
        else:
            logger.info("start eval...")
            device = torch.device("cuda")
            model.to(device)
            
        model.eval()
        total, acc = 0, 0
        save_file = dict()
        error_dict = dict()
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                probs = model(data)
                predicts = torch.argmax(probs, dim=1).cpu().tolist()
                targets = data['labels'].cpu().tolist()
                total += len(targets)
                results = [1 if predicts[j] == targets[j] else 0 for j in range(len(targets))]
                acc += sum(results)
                if save:
                    for j in range(len(targets)):
                        pred = get_keys_by_value(self.type_map, predicts[j])[0]
                        target = get_keys_by_value(self.type_map, targets[j])[0]
                        save_file[data['qids'][j]] = {
                            "predict": pred,
                            "target":  target,
                        }
                        if predicts[j] != targets[j]:
                            key_str = f'{target}->{pred}'
                            if key_str not in error_dict.keys():
                                error_dict[key_str] = 1
                            else:
                                error_dict[key_str] += 1
        if save:
            with open(os.path.join('./eval_result', 'error.json'), 'w') as f:
                json.dump(error_dict, f, indent=4)
            with open(os.path.join('./eval_result', 'classify_result.json'), 'w') as f:
                json.dump(save_file, f, indent=4)
        # logger.info(f"Total golds: {total}")
        return acc / total


    def test(self, ckpt_path = None, per_device_test_batch_size = 8):
        if ckpt_path == None and self.args.ckpt_for_test==None:
            raise ValueError("ckpt_path is empty")
        
        model = self.model
        device = torch.device("cuda")
        model.to(device)
        
        logger = create_logger("Testing", log_file=os.path.join(self.args.output_dir, self.args.log_file))
        if self.test_set == None:
            self.test_set = ClassifyDataset(tokenizer=self.tokenizer,type_fn=self.type_fn, type_map=self.type_map,
                                            data_path=self.args.data_path, split='test')

        loader = DataLoader(self.test_set, batch_size=per_device_test_batch_size, 
                                 collate_fn=lambda x: collate(x, self.tokenizer, self.args.max_length, test=True))

        model_load_path = ckpt_path if ckpt_path is not None else self.args.ckpt_for_test
        logger.info(f"loading trained parameters from {model_load_path}")
        model.load_state_dict(torch.load(model_load_path))
        logger.info("start infer....")
        model.eval()
        save_file = dict()
        total = 0
        with torch.no_grad():
            for i, data in enumerate(tqdm(loader)):
                probs = model(data)
                predicts = torch.argmax(probs, dim=1).cpu().tolist()
                total += len(predicts)
                for j in range(len(predicts)):
                    save_file[data['qids'][j]] = {
                        "predict": get_keys_by_value(self.type_map, predicts[j])[0],
                    }

        with open(self.args.test_out_path, 'w') as f:
            json.dump(save_file, f, indent=4)

        print(f"Total: {total}")


import json
import os
import math
import torch
import wandb
from tqdm import tqdm
import torch.nn.functional as F
from abc import abstractmethod, ABC
from torch.optim.adamw import AdamW
from typing import List, Dict
from torch.utils.data import (
    DataLoader,
    WeightedRandomSampler
)
from torch.optim.lr_scheduler import (
    CosineAnnealingLR,
    ReduceLROnPlateau,
    StepLR,
    CosineAnnealingWarmRestarts,
    PolynomialLR,
    LambdaLR
)
from transformers import (
    AutoTokenizer,
    AutoModel,
    AutoConfig,
    get_linear_schedule_with_warmup
)
from .dataset import RetrieveDataset, RetrieveRowDataset
from .utils import get_recall, right_pad_sequences, get_training_args, fix_seed
from .model import Retriever


class RetrieverTrainer(ABC):
    def __init__(self):
        self.training_args = get_training_args()
        fix_seed(self.training_args.random_seed)
        self.encoder = AutoModel.from_pretrained(
            self.training_args.encoder_path,
            trust_remote_code=True
        )
        self.config = AutoConfig.from_pretrained(self.training_args.encoder_path)
        self.model = Retriever(encoder=self.encoder, loss_fn_type=self.training_args.loss_fn_type,
                               input_dim=self.config.hidden_size).cuda()
        if self.training_args.ckpt_for_test is not None:
            self.model.load_state_dict(torch.load(self.training_args.ckpt_for_test))
        self.tokenizer = AutoTokenizer.from_pretrained(self.training_args.encoder_path, trust_remote_code=True)
        self.train_set = None
        self.val_set = None
        self.test_set = None
        self.optimizer = AdamW(self.model.parameters(), lr=self.training_args.learning_rate)
        if self.training_args.train_path:
            self.train_set = self.read_data(self.training_args.train_path)
            self.train_set = [self.data_proc(one) for one in self.train_set]
            if self.training_args.train_mode == "sample":
                self.train_set = RetrieveDataset(self.train_set)
            else:
                self.train_set = RetrieveRowDataset(self.train_set)
        if self.training_args.val_path:
            self.val_set = self.read_data(self.training_args.val_path)
            self.val_set = [self.data_proc(one) for one in self.val_set]
            self.val_set = RetrieveDataset(self.val_set)
        self.scheduler = None

    def get_scheduler(self):
        total_steps = len(self.train_set) * self.training_args.num_train_epochs / self.training_args.per_device_train_batch_size
        if self.training_args.lr_scheduler_type == "linear":
            return get_linear_schedule_with_warmup(
                self.optimizer,
                num_warmup_steps=self.training_args.warmup_steps,
                num_training_steps=total_steps
            )
        elif self.training_args.lr_scheduler_type == "cosine":
            return CosineAnnealingLR(
                self.optimizer,
                T_max=total_steps,
                eta_min=0,
                last_epoch=-1
            )
        elif self.training_args.lr_scheduler_type == "cosine_with_restarts":
            return CosineAnnealingWarmRestarts(
                self.optimizer,
                T_0=total_steps,
                T_mult=1,
                eta_min=0,
                last_epoch=-1
            )
        elif self.training_args.lr_scheduler_type == "polynomial":
            return PolynomialLR(
                self.optimizer,
                last_epoch=-1
            )
        elif self.training_args.lr_scheduler_type == "constant":
            return StepLR(
                self.optimizer,
                step_size=self.training_args.logging_steps,
                gamma=1
            )
        elif self.training_args.lr_scheduler_type == "constant_with_warmup":
            return get_linear_schedule_with_warmup(
                self.optimizer,
                num_warmup_steps=self.training_args.warmup_steps,
                num_training_steps=total_steps,
                last_epoch=-1
            )
        elif self.training_args.lr_scheduler_type == "inverse_sqrt":
            return LambdaLR(
                self.optimizer,
                lr_lambda=lambda epoch: 1 / math.sqrt(epoch),
                last_epoch=-1
            )
        elif self.training_args.lr_scheduler_type == "reduce_lr_on_plateau":
            return ReduceLROnPlateau(
                self.optimizer,
                mode='min',
                factor=0.1,
                patience=10
            )
        else:
            raise NotImplementedError

    def test_iterator(self):
        if self.training_args.test_path:
            self.test_set = self.read_data(self.training_args.test_path)
            self.test_set = [self.data_proc(one) for one in self.test_set]
            self.test_set = RetrieveDataset(self.test_set)
        else:
            raise ValueError(f'Unsupported test_path: {self.training_args.test_path}')

        test_loader = DataLoader(self.test_set,
                                 batch_size=1,
                                 num_workers=self.training_args.dataloader_num_workers,
                                 pin_memory=self.training_args.dataloader_pin_memory,
                                 shuffle=False,
                                 collate_fn=self.test_collate_fn
                                 )
        for index, one in enumerate(tqdm(test_loader, desc="infer stage:", leave=False)):
            with torch.no_grad():
                output = self.model(one)
                if self.training_args.loss_fn_type == "CE":
                    output = F.softmax(output, dim=1)[:, 1]
                elif self.training_args.loss_fn_type == "BCE":
                    pass
                else:
                    raise NotImplementedError
                _, index = torch.sort(output, dim=0, descending=True)
                output[index[:self.training_args.top_n_for_test]] = 1
                output[index[self.training_args.top_n_for_test:]] = 0
                evidence = torch.nonzero(output.reshape(-1) == 1).reshape(-1).long().tolist()
                yield evidence

    def eval(self, epoch: int = 0, eval_wandb: bool = False):
        if self.val_set is None:
            raise ValueError(f'Unsupported train_set: {self.val_set}')
        val_loader = DataLoader(self.val_set,
                                batch_size=self.training_args.per_device_eval_batch_size,
                                num_workers=self.training_args.dataloader_num_workers,
                                pin_memory=self.training_args.dataloader_pin_memory,
                                shuffle=False,
                                collate_fn=self.train_collate_fn_for_eval
                                )
        self.model.eval()
        val_loss = 0
        val_recall = 0
        val_cor = 0
        with torch.no_grad():
            for val_inputs in tqdm(val_loader, desc="Validation:", leave=False):
                outputs = self.model(val_inputs)
                if self.training_args.loss_fn_type == "CE":
                    _, pred = torch.max(outputs, 1)
                else:
                    a = torch.ones_like(outputs)
                    b = torch.zeros_like(outputs)
                    pred = torch.where(outputs > 0.5, a, b)
                loss = self.model.loss
                val_loss += (loss.sum().item() / outputs.shape[0])
                val_recall += get_recall(outputs, val_inputs["labels"], self.training_args.top_n_for_eval)
                val_cor += ((pred == val_inputs["labels"]).sum().item() / outputs.shape[0])

            epoch_loss = val_loss / len(self.val_set)
            epoch_acc = val_cor / len(self.val_set)
            epoch_recall = val_recall / len(val_loader)
            print('Validation - Epoch: {}  Loss: {:.6f}  Recall: {:.6f}  Acc: {:.6f}%'.format(
                epoch + 1, epoch_loss, epoch_recall, epoch_acc * 100
            ))
            if self.training_args.use_wandb and eval_wandb:
                wandb.log({'eval/loss': epoch_loss, 'eval/Acc': epoch_acc, 'eval/Recall': epoch_recall})

    def save(self, epoch, step):
        print("Saving......")
        output_directory = self.training_args.output_dir
        if not os.path.exists(output_directory):
            os.makedirs(output_directory)
        model_name = 'epoch{}_step{}.pt'.format(epoch + 1, step)
        save_path = os.path.join(self.training_args.output_dir, model_name)
        torch.save(self.model.state_dict(), save_path)

    def train(self):
        if self.train_set is None:
            raise ValueError(f'Unsupported train_set: {self.train_set}')
        if self.training_args.use_wandb:
            wandb.init(project="Retriever")
        self.scheduler = self.get_scheduler()
        if self.training_args.train_mode == "row" and self.training_args.use_sampler:
            train_sampler = WeightedRandomSampler(weights=self.train_set.make_weights(),
                                                  num_samples=len(self.train_set), replacement=True)
            train_loader = DataLoader(self.train_set,
                                      batch_size=self.training_args.per_device_train_batch_size,
                                      num_workers=self.training_args.dataloader_num_workers,
                                      pin_memory=self.training_args.dataloader_pin_memory,
                                      shuffle=False,
                                      sampler=train_sampler,
                                      collate_fn=self.train_collate_fn
                                      )
        else:
            train_loader = DataLoader(self.train_set,
                                      batch_size=self.training_args.per_device_train_batch_size,
                                      num_workers=self.training_args.dataloader_num_workers,
                                      pin_memory=self.training_args.dataloader_pin_memory,
                                      shuffle=True,
                                      collate_fn=self.train_collate_fn
                                      )

        step = 0
        running_loss = 0
        running_recall = 0
        cor = 0
        num_sample = 0
        for epoch in range(int(self.training_args.num_train_epochs)):
            for inputs in tqdm(train_loader, desc="Training:", leave=False):
                torch.cuda.empty_cache()
                self.optimizer.zero_grad()
                self.scheduler.step()
                self.model.train()
                outputs = self.model(inputs)
                if self.training_args.loss_fn_type == "CE":
                    _, pred = torch.max(outputs, 1)
                else:
                    a = torch.ones_like(outputs)
                    b = torch.zeros_like(outputs)
                    pred = torch.where(outputs > 0.5, a, b)
                loss = self.model.loss
                loss.backward()
                self.optimizer.step()
                running_loss += loss.sum().item()
                running_recall += get_recall(outputs, inputs["labels"], self.training_args.top_n_for_eval)
                cor += (pred == inputs["labels"]).sum().item()
                step += 1
                num_sample += outputs.shape[0]
                if step % self.training_args.logging_steps == 0:
                    step_loss = running_loss / num_sample
                    step_acc = cor / num_sample
                    step_recall = running_recall / self.training_args.logging_steps
                    running_loss = 0
                    running_recall = 0
                    cor = 0
                    num_sample = 0
                    print(
                        'Training - Step: {}  Loss: {:.6f}  Recall: {:.6f}  Acc: {:.6f}%  Learning Rate: {:.9f}'.format(
                            step, step_loss, step_recall, step_acc * 100, self.scheduler.get_last_lr()[0]
                        ))
                    if self.training_args.use_wandb:
                        wandb.log({'train/epoch': epoch + 1, 'train/step': step, 'train/loss': step_loss,
                                   'train/lr': self.scheduler.get_last_lr()[0], 'train/Acc': step_acc,
                                   'train/Recall': step_recall})

                if step % self.training_args.save_steps == 0:
                    self.save(epoch, step)

                if self.val_set and step % self.training_args.eval_steps == 0:
                    self.eval(epoch, True)
        self.save(int(self.training_args.num_train_epochs) - 1, step)

    @abstractmethod
    def read_data(self, data_path: str) -> List[Dict]:
        """
        Read the data from the specified data path.

        :param data_path: The path to the data.
        :return: A list of dictionaries representing the raw data.
        Each dictionary should contain the data for one instance.
        Example:
        [
            {"id": "1", "question": "What is this?", "rows": ["row1 text", "row2 text"], "labels": [0, 1], ...},
            {"id": "2", "question": "Another question", "rows": ["row1 text"], "labels": [1], ...},
            ...
        ]
        """
        pass

    @abstractmethod
    def data_proc(self, instance) -> Dict:
        """
        Preprocess one instance of the dataset into the required format.

        :param instance: One instance of the dataset.
        :return: A dictionary containing the preprocessed instance.
        The dictionary should have the following keys:
        - "id": str (unique identifier for the instance)
        - "question": str (the question text)
        - "rows": List[str] (a list of text rows)
        - "labels": List[int] (a list of labels)
        Example:
        {
            "id": "1",
            "question": "What is this?",
            "rows": ["row1 text", "row2 text"],
            "labels": [0, 1]
        }
        """
        pass

    def train_collate_fn_for_eval(self, batch):
        inputs = {
            "input_ids": [],
            "attention_mask": [],
            "labels": []
        }
        for one in batch:
            question = one["question"]
            rows = one["rows"]
            inputs["labels"] += one["labels"]

            for row in rows:
                row_ids = self.tokenizer.encode(question + self.tokenizer.sep_token + row)
                inputs["input_ids"].append(row_ids)
                inputs["attention_mask"].append([1] * len(row_ids))
        inputs["input_ids"] = right_pad_sequences(inputs["input_ids"], padding_value=self.tokenizer.pad_token_id,
                                                  max_len=-1, truncation=self.training_args.truncation)
        inputs["attention_mask"] = right_pad_sequences(inputs["attention_mask"], padding_value=0, max_len=-1,
                                                       truncation=self.training_args.truncation)
        inputs["labels"] = torch.LongTensor(inputs["labels"]).cuda()
        return inputs

    def train_collate_fn(self, batch):
        inputs = {
            "input_ids": [],
            "attention_mask": [],
            "labels": []
        }
        for one in batch:
            question = one["question"]
            rows = one["rows"]
            inputs["labels"] += one["labels"]
            if self.training_args.train_mode == "sample":
                for row in rows:
                    row_ids = self.tokenizer.encode(question + self.tokenizer.sep_token + row)
                    inputs["input_ids"].append(row_ids)
                    inputs["attention_mask"].append([1] * len(row_ids))
            else:
                row_ids = self.tokenizer.encode(question + self.tokenizer.sep_token + rows)
                inputs["input_ids"].append(row_ids)
                inputs["attention_mask"].append([1] * len(row_ids))
        inputs["input_ids"] = right_pad_sequences(inputs["input_ids"], padding_value=self.tokenizer.pad_token_id,
                                                  max_len=-1, truncation=self.training_args.truncation)
        inputs["attention_mask"] = right_pad_sequences(inputs["attention_mask"], padding_value=0, max_len=-1,
                                                       truncation=self.training_args.truncation)
        inputs["labels"] = torch.LongTensor(inputs["labels"]).cuda()
        return inputs

    def test_collate_fn(self, batch):
        inputs = {
            "input_ids": [],
            "attention_mask": [],
            "labels": None
        }
        for one in batch:
            question = one["question"]
            rows = one["rows"]
            for row in rows:
                row_ids = self.tokenizer.encode(question + self.tokenizer.sep_token + row)
                inputs["input_ids"].append(row_ids)
                inputs["attention_mask"].append([1] * len(row_ids))
        inputs["input_ids"] = right_pad_sequences(inputs["input_ids"], padding_value=self.tokenizer.pad_token_id,
                                                  max_len=-1, truncation=self.training_args.truncation)
        inputs["attention_mask"] = right_pad_sequences(inputs["attention_mask"], padding_value=0, max_len=-1,
                                                       truncation=self.training_args.truncation)
        return inputs


class MultiHierttTrainer(RetrieverTrainer):
    def read_data(self, data_path: str) -> List[Dict]:
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        return data

    def data_proc(self, instance) -> Dict:
        rows = instance["paragraphs"]
        labels = [0] * len(instance["paragraphs"])
        if len(instance["qa"]["text_evidence"]):
            for text_evidence in instance["qa"]["text_evidence"]:
                labels[text_evidence] = 1
        for k, v in instance["table_description"].items():
            rows.append(v)
            labels.append(1 if k in instance["qa"]["table_evidence"] else 0)
        return {
            "id": instance["uid"],
            "question": instance["qa"]["question"],
            "rows": rows,
            "labels": labels
        }

    def infer(self):
        test_data = self.read_data(self.training_args.test_path)
        for one, pred in zip(test_data, self.test_iterator()):
            text_pred = pred[:len(one["paragraphs"])]
            table_pred = pred[len(one["paragraphs"]):]
            one["qa"]["text_evidence"] = [i for i, x in enumerate(text_pred) if x == 1]
            keys = list(one["table_description"].keys())
            one["qa"]["text_evidence"] = [key for i, key in enumerate(keys) if table_pred[i] == 1]

            if not os.path.isfile(self.training_args.test_out_path):
                with open(self.training_args.test_out_path, 'w') as file:
                    json.dump([], file)
            with open(self.training_args.test_out_path, 'r+') as file:
                file_data = json.load(file)
                file_data.append(one)
                file.seek(0)
                json.dump(file_data, file, indent=4)


class FinQATrainer(RetrieverTrainer):
    @staticmethod
    def remove_space(text_in):
        res = []

        for tmp in text_in.split(" "):
            if tmp != "":
                res.append(tmp)

        return " ".join(res)

    def table_row_to_text(self, header, row):
        res = ""

        if header[0]:
            res += (header[0] + " ")

        for head, cell in zip(header[1:], row[1:]):
            res += ("the " + row[0] + " of " + head + " is " + cell + " ; ")

        res = self.remove_space(res)
        return res.strip()

    def read_data(self, data_path: str) -> List[Dict]:
        data = json.load(
            open(data_path, 'r', encoding='utf-8')
        )
        return data

    def data_proc(self, instance) -> Dict:
        gold_inds = instance["qa"]["gold_inds"]
        pre_text = instance["pre_text2"]
        post_text = instance["post_text2"]
        rows = pre_text + post_text
        labels = [0] * (len(rows))
        for key in gold_inds:
            if "text_" in key:
                text_id = int(key.replace("text_", ""))
                if text_id < len(labels):
                    labels[text_id] = 1

        table = instance["table"]
        for idx, row in enumerate(table):
            row_description = self.table_row_to_text(table[0], table[idx])
            rows.append(row_description)
            label_key = "table_" + str(idx)
            labels.append(1 if label_key in gold_inds else 0)
        return {
                "id": instance["id"],
                "question": instance["qa"]["question"],
                "rows": rows,
                "labels": labels
            }

    def infer(self):
        test_data = self.read_data(self.training_args.test_path)
        for one, pred in zip(test_data, self.test_iterator()):
            texts = one["pre_text2"] + one["post_text2"]
            text_pred = pred[:len(texts)]
            table_pred = pred[len(texts):]
            one["qa"]["ann_text_rows"] = [i for i, x in enumerate(text_pred) if x == 1]
            one["qa"]["ann_table_rows"] = [i for i, x in enumerate(table_pred) if x == 1]
            for i in one["qa"]["ann_text_rows"]:
                one["qa"]["gold_inds"][f"text_{i}"] = texts[i]

            for i in one["qa"]["ann_table_rows"]:
                one["qa"]["gold_inds"][f"table_{i}"] = self.table_row_to_text(one["table"][0], one["table"][i])

            if not os.path.isfile(self.training_args.test_out_path):
                with open(self.training_args.test_out_path, 'w') as file:
                    json.dump([], file)
            with open(self.training_args.test_out_path, 'r+') as file:
                file_data = json.load(file)
                file_data.append(one)
                file.seek(0)
                json.dump(file_data, file, indent=4)


class CompAQTTrainer(FinQATrainer):
    def data_proc(self, instance) -> Dict:
        gold_inds = instance["qa"]["gold_inds"]
        pre_text = instance["pre_text"]
        post_text = instance["post_text"]
        rows = pre_text + post_text  # 所有备选文本  序号是0-index的和
        labels = [0] * (len(rows))
        for key in gold_inds:  # match the text
            if "text_" in key:
                text_id = int(key.replace("text_", ""))
                labels[text_id] = 1

        table = instance["table"]  # finqa只有一个表格 对每行进行table_discr的操作
        for idx, row in enumerate(table):  # match the row of table
            row_description = self.table_row_to_text(table[0], table[idx])
            rows.append(row_description)
            label_key = "table_" + str(idx)
            labels.append(1 if label_key in gold_inds else 0)
        return {
                "id": instance["id"],
                "question": instance["qa"]["question"],
                "rows": rows,
                "labels": labels
            }

    def infer(self):
        test_data = self.read_data(self.training_args.test_path)
        for one, pred in zip(test_data, self.test_iterator()):
            texts = one["pre_text"] + one["post_text"]
            text_pred = pred[:len(texts)]
            table_pred = pred[len(texts):]
            one["qa"]["ann_text_rows"] = [i for i, x in enumerate(text_pred) if x == 1]
            one["qa"]["ann_table_rows"] = [i for i, x in enumerate(table_pred) if x == 1]
            for i in one["qa"]["ann_text_rows"]:
                one["qa"]["gold_inds"][f"text_{i}"] = texts[i]

            for i in one["qa"]["ann_table_rows"]:
                one["qa"]["gold_inds"][f"table_{i}"] = self.table_row_to_text(one["table"][0], one["table"][i])

            if not os.path.isfile(self.training_args.test_out_path):
                with open(self.training_args.test_out_path, 'w') as file:
                    json.dump([], file)
            with open(self.training_args.test_out_path, 'r+') as file:
                file_data = json.load(file)
                file_data.append(one)
                file.seek(0)
                json.dump(file_data, file, indent=4)



import os
import json
from .args import RunnningArguments, get_running_args
from transformers import HfArgumentParser
import time
import json
import argparse
import copy
import os

from typing import List
import platform
import multiprocessing

from .qa import openai_qa
from .qa.openai_qa import Generator
from .utils import load_data_split, parse_ans, get_type
from utils.retriever import Retriever

class MMQA_Runner:
    def __init__(self, api_keys=None) -> None:
        self.args = get_running_args()
        # Build paths
        args = self.args
        os.makedirs(args.save_dir, exist_ok=True)

        # Load dataset
        
        self.dataset = load_data_split(dataset_to_load='mmqa',split=args.split, data_dir=args.data_path) # 从本地加载加载数据集

        # Load openai keys
        if api_keys==None and args.api_keys_file is not None:
            with open(args.api_keys_file, 'r') as f:
                self.keys = [line.strip() for line in f.readlines()]
        elif api_keys!=None:
            self.keys = api_keys
        else:
            raise ValueError('no api keys provided')
    
    def worker_qa(
        self, 
        pid: int,
        args,
        generator: Generator,
        g_eids: List,
        dataset
    ):
        """
        A worker process for end to end qa.
        """
        retriever = Retriever(args)
        g_dict = dict()
        built_few_shot_prompts = []
        with open(os.path.join(args.template_path, args.prompt_file), 'r') as f:
            build_cfg = json.load(f)

        try:
            for g_eid in g_eids:
                try:
                    g_data_item = dataset[g_eid]
                    if args.oracle_retriever:
                        new_passages = {"id": [], "title": [], "text": [], "url": []}
                        new_images = {"id": [], "title": [], "pic": [], "url": [], "path": []}
                        # Get the retrieved passages
                        for _id, title, text, url in zip(g_data_item['passages']['id'],
                                                    g_data_item['passages']['title'],
                                                    g_data_item['passages']['text'],
                                                    g_data_item['passages']['url']):
                            if _id in g_data_item['supporting_context']['doc_id']:
                                new_passages["id"].append(_id)
                                new_passages["title"].append(title)
                                new_passages["text"].append(text)
                                new_passages["url"].append(url)
                        
                        # Get the retrieved images
                        for _id, title, pic, url, path in zip(g_data_item['images']['id'],
                                                        g_data_item['images']['title'],
                                                        g_data_item['images']['pic'],
                                                        g_data_item['images']['url'],
                                                        g_data_item['images']['path']):
                            if _id in g_data_item['supporting_context']['doc_id']:
                                new_images["id"].append(_id)
                                new_images["title"].append(title)
                                new_images["pic"].append(pic)
                                new_images["url"].append(url)
                                new_images["path"].append(path)
                    else:
                        new_passages, new_images, qtype = retriever.retrieve(g_data_item, g_eid, g_data_item['id'])
                    if args.oracle_classifier:
                        qtype = get_type(g_data_item['type'])

                    g_data_item['passages'] = new_passages
                    g_data_item['images'] = new_images
                    g_dict[g_eid] = {
                        'generations': '',
                        'ori_data_item': copy.deepcopy(g_data_item)
                    }

                    prompt_cfg = build_cfg[qtype].copy()
                    
                    # Build prompt
                    generate_prompt = generator.build_generate_prompt(
                        data_item=g_data_item,
                        qtype=qtype,
                        cot=prompt_cfg['cot']
                    )

                    ok = False
                    n_shots = prompt_cfg['n-shot']
                    while not ok:

                        built_few_shot_prompts = []
                        few_shot_prompt = generator.build_few_shot_prompt_from_file(
                            file_path=prompt_cfg['file'],
                            n_shots=n_shots,
                            qtype = qtype
                        )
                        
                        prompt = few_shot_prompt + "\n\n\n" + generate_prompt # 手造的 few shot prompt + 需要提交的 prompt

                        print(f"Process#{pid}: Building prompt for eid#{g_eid}, original_id#{g_data_item['id']}")
                        
                        built_few_shot_prompts.append((g_eid, prompt))

                        print(f"Process#{pid}: Prompts ready with {len(built_few_shot_prompts)} parallels. Run {args.engine} API.")

                        ok, response_dict = generator.generate_one_pass(
                            prompts=built_few_shot_prompts,
                            args = prompt_cfg, 
                            verbose=args.verbose
                        )

                        if not ok:
                            assert n_shots >= 0, "n_shots should be no less than 0."
                            n_shots -= 1
                    for eid, g_pairs in response_dict.items():
                        g_pairs = sorted(g_pairs, key=lambda x: x[-1], reverse=True)
                        pred_ans = g_pairs[0][0]

                        for g_pair in g_pairs:
                            status, ans_text = parse_ans(g_pair[0],qtype, cot=prompt_cfg['cot'])
                            if status:
                                pred_ans = ans_text
                                break
                        g_dict[eid]['generations'] = pred_ans.strip()
                    built_few_shot_prompts = []
                except Exception as e:
                    print(f"Process#{pid}: eid#{g_eid}, qid#{g_data_item['id']} generation error: {e}")
                    break
        except KeyboardInterrupt:
            g_dict.popitem()
            print(f"Process#{pid}: Keyboard interrupt. Saving the generations.")
        return g_dict

    def predict(self):
        start_time = time.time()
        args = self.args
        dataset = self.dataset
        keys = self.keys
        # End To End QA
        generator = openai_qa.Generator(args, keys) # 调用 LLM api 的
        generate_eids = list(range(len(dataset)))
        generate_eids_group = [[] for _ in range(args.n_processes)]
        for g_eid in generate_eids:
            generate_eids_group[int(g_eid) % args.n_processes].append(g_eid) # 每个进程处理的数据集, 按剩余系分
        print('\n******* End To End QA *******')
        g_dict = dict()
        worker_results = []
        if args.n_processes > 1:
            pool = multiprocessing.Pool(processes=args.n_processes)
            for pid in range(args.n_processes): # 每个 api_key 对应一个进程
                worker_results.append(pool.apply_async(self.worker_qa, args=(
                    pid,
                    args,
                    generator,
                    generate_eids_group[pid],
                    dataset,
                )))

            # Merge results
            for r in worker_results:
                worker_g_dict = r.get()
                g_dict.update(worker_g_dict)
            pool.close()
            pool.join()
            # Save results
            # "_".join(["{}={}".format(k, str(args.__dict__[k])) for k in args.__dict__ if k not in ['api_keys_file', 'prompt_file', 'save_dir', 'stop_tokens']])
            save_file_name = f'mmhqa-icl_mmqa_{args.split}.json'
            with open(os.path.join(args.save_dir, save_file_name), 'w') as f:
                json.dump(g_dict, f, indent=4)
        else:
            """
            for resume qa with n_processes=1
            """
            if args.save_file_name is None:
                save_file_name = f'mmhqa-icl_mmqa_{args.split}.json'
            else :
                save_file_name = args.save_file_name

            begin_eid = 0
            end_eid = len(generate_eids)

            g_dict = dict()
            if args.resume:
                try:
                    with open(os.path.join(args.save_dir, save_file_name), 'r') as f:
                        g_dict = json.load(f)
                        begin_eid = len(g_dict)
                except Exception as e:
                    print(e)
                    print(f"Resume failed")
                    exit(0)

            print("begin end to end qa begin_eid: ", begin_eid, "end_eid: ", end_eid, "total: ", len(generate_eids)) 

            n_g_dict = self.worker_qa(
                0,
                args,
                generator,
                generate_eids[begin_eid:end_eid], # 从上次处理的位置开始
                dataset,
            )
            g_dict.update(n_g_dict)

            with open(os.path.join(args.save_dir, save_file_name), 'w') as f:
                json.dump(g_dict, f, indent=4)

        print(f"Elapsed time: {time.time() - start_time}")


#coding=utf8
import dgl, math, torch

def src_dot_dst(src_field, dst_field, out_field):
    def func(edges):
        return {out_field: (edges.src[src_field] * edges.dst[dst_field]).sum(-1, keepdim=True)}

    return func

def src_sum_edge_mul_dst(src_field, dst_field, e_field, out_field):
    def func(edges):
        return {out_field: ((edges.src[src_field] + edges.data[e_field]) * edges.dst[dst_field]).sum(-1, keepdim=True)}

    return func

def scaled_exp(field, scale_constant):
    def func(edges):
        # clamp for softmax numerical stability
        return {field: torch.exp((edges.data[field] / scale_constant).clamp(-10, 10))}

    return func

def src_sum_edge_mul_edge(src_field, e_field1, e_field2, out_field):
    def func(edges):
        return {out_field: (edges.src[src_field] + edges.data[e_field1]) * edges.data[e_field2]}

    return func

def div_by_z(in_field, norm_field, out_field):
    def func(nodes):
        return {out_field: nodes.data[in_field] / nodes.data[norm_field]}

    return func


import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import numpy as np
from transformers import BartTokenizer, BartForConditionalGeneration
from transformers import RobertaTokenizer, RobertaModel, BertTokenizer, BertModel, BertForQuestionAnswering
import json
from tqdm import tqdm
import random
from pathlib import Path
import logging
import os, sys, math
import torch.nn.functional as F
from transformers import get_linear_schedule_with_warmup, AdamW


def gelu(x):
    return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))
class FFNLayer(nn.Module):
    def __init__(self, input_dim, intermediate_dim, output_dim, dropout, layer_norm=True):
        super(FFNLayer, self).__init__()
        self.fc1 = nn.Linear(input_dim, intermediate_dim)
        if layer_norm:
            self.ln = nn.LayerNorm(intermediate_dim)
        else:
            self.ln = None
        self.dropout_func = nn.Dropout(dropout)
        self.fc2 = nn.Linear(intermediate_dim, output_dim)

    def forward(self, input):
        inter = self.fc1(self.dropout_func(input))
        inter_act = gelu(inter)
        if self.ln:
            inter_act = self.ln(inter_act)
        return self.fc2(inter_act)    


class ReadModel(nn.Module):
    def __init__(self, bert_model):
        super(ReadModel, self).__init__()
        self.bert_model = bert_model
        self.hidden_size = self.bert_model.embeddings.word_embeddings.embedding_dim
        self.start_pre = FFNLayer(self.hidden_size, self.hidden_size, 1, 0.2)
        self.end_pre = FFNLayer(self.hidden_size, self.hidden_size, 1, 0.2)

    def forward(self, data):
        inputs = {"input_ids": data['input_ids'], "attention_mask": data['input_mask']}
        output = self.bert_model(**inputs)[0]
        start_logits = self.start_pre(output).squeeze(-1)
        end_logits = self.end_pre(output).squeeze(-1)
        return start_logits, end_logits
    
